from libcpp cimport bool
from weakref import ref
import atexit

cdef extern from 'gst/gst.h':
    ctypedef void *GstPipeline
    ctypedef void *GstElement
    ctypedef void *GstBus
    ctypedef void *GstPad
    ctypedef void *GstSample
    ctypedef void *GstBin
    ctypedef void (*appcallback_t)(void *, int, int, char *, int)
    ctypedef void (*buscallback_t)(void *, GstMessage *)
    ctypedef unsigned int guint
    ctypedef unsigned long gulong
    ctypedef void *gpointer
    ctypedef char const_gchar 'const gchar'
    ctypedef long int gint64
    ctypedef unsigned long long GstClockTime
    ctypedef int gboolean

    ctypedef enum GstState:
        GST_STATE_VOID_PENDING
        GST_STATE_NULL
        GST_STATE_READY
        GST_STATE_PAUSED
        GST_STATE_PLAYING

    ctypedef enum GstFormat:
        GST_FORMAT_TIME

    ctypedef enum GstSeekFlags:
        GST_SEEK_FLAG_KEY_UNIT
        GST_SEEK_FLAG_FLUSH

    ctypedef enum GstStateChangeReturn:
        pass

    ctypedef struct GError:
        int code
        char *message

    ctypedef enum GstMessageType:
        GST_MESSAGE_EOS
        GST_MESSAGE_ERROR
        GST_MESSAGE_WARNING
        GST_MESSAGE_INFO

    ctypedef struct GstMessage:
        GstMessageType type

    int GST_SECOND
    bool gst_init_check(int *argc, char ***argv, GError **error)
    bool gst_is_initialized()
    void gst_deinit()
    void gst_version(guint *major, guint *minor, guint *micro, guint *nano)
    GstElement *gst_element_factory_make(const_gchar *factoryname, const_gchar *name)
    bool gst_bin_add(GstBin *bin, GstElement *element)
    bool gst_bin_remove(GstBin *bin, GstElement *element)
    void gst_object_unref(void *pointer) nogil
    GstElement *gst_pipeline_new(const_gchar *name)
    void gst_bus_enable_sync_message_emission(GstBus *bus)
    GstBus *gst_pipeline_get_bus(GstPipeline *pipeline)
    GstStateChangeReturn gst_element_get_state(
            GstElement *element, GstState *state, GstState *pending,
            GstClockTime timeout) nogil
    GstStateChangeReturn gst_element_set_state(
            GstElement *element, GstState state) nogil
    void g_signal_emit_by_name(gpointer instance, const_gchar *detailed_signal,
            void *retvalue)
    void g_error_free(GError *error)
    bool gst_element_query_position(
            GstElement *element, GstFormat format, gint64 *cur) nogil
    bool gst_element_query_duration(
            GstElement *element, GstFormat format, gint64 *cur) nogil
    bool gst_element_seek_simple(
            GstElement *element, GstFormat format,
            GstSeekFlags seek_flags, gint64 seek_pos) nogil
    void gst_message_parse_error(
            GstMessage *message, GError **gerror, char **debug)
    void gst_message_parse_warning(
            GstMessage *message, GError **gerror, char **debug)
    void gst_message_parse_info(
            GstMessage *message, GError **gerror, char **debug)

cdef extern from '_gstplayer.h':
    void g_object_set_void(GstElement *element, char *name, void *value)
    void g_object_set_double(GstElement *element, char *name, double value) nogil
    void g_object_set_caps(GstElement *element, char *value)
    void g_object_set_int(GstElement *element, char *name, int value)
    gulong c_appsink_set_sample_callback(GstElement *appsink,
            appcallback_t callback, void *userdata)
    void c_appsink_pull_preroll(GstElement *appsink,
            appcallback_t callback, void *userdata) nogil
    gulong c_bus_connect_message(GstBus *bus,
            buscallback_t callback, void *userdata)
    void c_signal_disconnect(GstElement *appsink, gulong handler_id)
    void c_glib_iteration(int count)


#
# prevent gstreamer crash when some player are still working.
#

cdef list _instances = []

def _on_player_deleted(wk):
    if wk in _instances:
        _instances.remove(wk)

@atexit.register
def gst_exit_clean():
    # XXX don't use a stop() method or anything that change the state of the
    # element without releasing the GIL. Otherwise, we might have a deadlock due
    # to GIL in appsink callback + GIL already locked here.
    for wk in _instances:
        player = wk()
        if player:
            player.unload()


class GstPlayerException(Exception):
    pass


cdef void _on_appsink_sample(
        void *c_player, int width, int height,
        char *data, int datasize) with gil:
    cdef GstPlayer player = <GstPlayer>c_player
    cdef bytes buf = data[:datasize]
    if player.sample_cb:
        player.sample_cb(width, height, buf)


cdef void _on_gstplayer_message(void *c_player, GstMessage *message) with gil:
    cdef GstPlayer player = <GstPlayer>c_player
    cdef GError *err = NULL
    if message.type == GST_MESSAGE_EOS:
        player.got_eos()
    elif message.type == GST_MESSAGE_ERROR:
        gst_message_parse_error(message, &err, NULL)
        player.message_cb('error', err.message)
        g_error_free(err);
    elif message.type == GST_MESSAGE_WARNING:
        gst_message_parse_warning(message, &err, NULL)
        player.message_cb('warning', err.message)
        g_error_free(err);
    elif message.type == GST_MESSAGE_INFO:
        gst_message_parse_info(message, &err, NULL)
        player.message_cb('info', err.message)
        g_error_free(err);
    else:
        pass

def _gst_init():
    if gst_is_initialized():
        return True
    cdef int argc = 0
    cdef char **argv = NULL
    cdef GError *error
    if not gst_init_check(&argc, &argv, &error):
        msg = 'Unable to initialize gstreamer: code={} message={}'.format(
                error.code, <bytes>error.message)
        raise GstPlayerException(msg)

def get_gst_version():
    cdef unsigned int major, minor, micro, nano
    gst_version(&major, &minor, &micro, &nano)
    return (major, minor, micro, nano)


def glib_iteration(int loop):
    c_glib_iteration(loop)


cdef class GstPlayer:
    cdef GstElement *pipeline
    cdef GstElement *playbin
    cdef GstElement *appsink
    cdef GstElement *fakesink
    cdef GstBus *bus
    cdef object uri, sample_cb, eos_cb, message_cb
    cdef gulong hid_sample, hid_message
    cdef object __weakref__

    def __cinit__(self, *args, **kwargs):
        self.pipeline = self.playbin = self.appsink = self.fakesink = NULL
        self.bus = NULL
        self.hid_sample = self.hid_message = 0

    def __init__(self, uri, sample_cb=None, eos_cb=None, message_cb=None):
        super(GstPlayer, self).__init__()
        self.uri = uri
        self.sample_cb = sample_cb
        self.eos_cb = eos_cb
        self.message_cb = message_cb
        _instances.append(ref(self, _on_player_deleted))

        # ensure gstreamer is init
        _gst_init()

    def __dealloc__(self):
        self.unload()

    cdef void got_eos(self):
        if self.eos_cb:
            self.eos_cb()

    def load(self):
        cdef bytes py_uri

        # if already loaded before, clean everything.
        if self.pipeline != NULL:
            self.unload()

        # create the pipeline
        self.pipeline = gst_pipeline_new(NULL)
        if self.pipeline == NULL:
            raise GstPlayerException('Unable to create a pipeline')

        self.bus = gst_pipeline_get_bus(<GstPipeline *>self.pipeline)
        if self.bus == NULL:
            raise GstPlayerException('Unable to get the bus from the pipeline')

        gst_bus_enable_sync_message_emission(self.bus)
        if self.eos_cb or self.message_cb:
            self.hid_message = c_bus_connect_message(
                    self.bus, _on_gstplayer_message, <void *>self)

        # instantiate the playbin
        self.playbin = gst_element_factory_make('playbin', NULL)
        if self.playbin == NULL:
            raise GstPlayerException(
                'Unable to create a playbin. Consider setting the environment variable '
                'GST_REGISTRY to a user accessible path, such as ~/registry.bin')

        gst_bin_add(<GstBin *>self.pipeline, self.playbin)

        # instantiate an appsink
        if self.sample_cb:
            self.appsink = gst_element_factory_make('appsink', NULL)
            if self.appsink == NULL:
                raise GstPlayerException('Unable to create an appsink')

            g_object_set_caps(self.appsink, 'video/x-raw,format=RGB')
            g_object_set_int(self.appsink, 'max-buffers', 5)
            g_object_set_int(self.appsink, 'drop', 1)
            g_object_set_int(self.appsink, 'sync', 1)
            g_object_set_int(self.appsink, 'qos', 1)
            g_object_set_void(self.playbin, 'video-sink', self.appsink)

        else:
            self.fakesink = gst_element_factory_make('fakesink', NULL)
            if self.fakesink == NULL:
                raise GstPlayerException('Unable to create a fakesink')

            g_object_set_void(self.playbin, 'video-sink', self.fakesink)

        # configure playbin
        g_object_set_int(self.pipeline, 'async-handling', 1)
        py_uri = <bytes>self.uri.encode('utf-8')
        g_object_set_void(self.playbin, 'uri', <char *>py_uri)

        # attach the callback
        # NOTE no need to create a weakref here, as we manage to grab/release
        # the reference of self in the set_sample_callback() method.
        if self.sample_cb:
            self.hid_sample = c_appsink_set_sample_callback(
                    self.appsink, _on_appsink_sample, <void *>self)

        # get ready!
        with nogil:
            gst_element_set_state(self.pipeline, GST_STATE_READY)

    def play(self):
        if self.pipeline != NULL:
            with nogil:
                gst_element_set_state(self.pipeline, GST_STATE_PLAYING)

    def stop(self):
        if self.pipeline != NULL:
            with nogil:
                gst_element_set_state(self.pipeline, GST_STATE_NULL)
                gst_element_set_state(self.pipeline, GST_STATE_READY)

    def pause(self):
        if self.pipeline != NULL:
            with nogil:
                gst_element_set_state(self.pipeline, GST_STATE_PAUSED)

    def unload(self):
        cdef GstState current_state, pending_state

        if self.appsink != NULL and self.hid_sample != 0:
            c_signal_disconnect(self.appsink, self.hid_sample)
            self.hid_sample = 0

        if self.bus != NULL and self.hid_message != 0:
            c_signal_disconnect(<GstElement *>self.bus, self.hid_message)
            self.hid_message = 0

        if self.pipeline != NULL:
            # the state changes are async. if we want to guarantee that the
            # state is set to NULL, we need to query it. We also put a 5s
            # timeout for safety, but normally, nobody should hit it.
            with nogil:
                gst_element_set_state(self.pipeline, GST_STATE_NULL)
                gst_element_get_state(self.pipeline, &current_state,
                        &pending_state, <GstClockTime>5e9)
            gst_object_unref(self.pipeline)

        if self.bus != NULL:
            gst_object_unref(self.bus)

        self.appsink = NULL
        self.bus = NULL
        self.pipeline = NULL
        self.playbin = NULL
        self.fakesink = NULL

    def set_volume(self, float volume):
        if self.playbin != NULL:
            # XXX we need to release the GIL, on linux, you might have a race
            # condition. When running, if pulseaudio is used, it might sent a
            # message when you set the volume, in the pulse audio thread
            # The message is received by our common sync-message, and try to get
            # the GIL, and block, because here we didn't release it.
            # 1. our thread get the GIL and ask pulseaudio to set the volume
            # 2. the pulseaudio thread try to sent a message, and wait for the
            #    GIL
            with nogil:
                g_object_set_double(self.playbin, 'volume', volume)

    def get_duration(self):
        cdef double duration
        with nogil:
            duration = <double>self._get_duration()
        if duration == -1:
            return -1
        return duration / float(GST_SECOND)

    def get_position(self):
        cdef double position
        with nogil:
            position = <double>self._get_position()
        if position == -1:
            return -1
        return position / float(GST_SECOND)

    def seek(self, float percent):
        with nogil:
            self._seek(percent)

    #
    # C-like API, that doesn't require the GIL
    #

    cdef gint64 _get_duration(self) nogil:
        cdef gint64 duration = -1
        cdef GstState state
        if self.playbin == NULL:
            return -1

        # check the state
        gst_element_get_state(self.pipeline, &state, NULL,
                <GstClockTime>GST_SECOND)

        # if we are already prerolled, we can read the duration
        if state == GST_STATE_PLAYING or state == GST_STATE_PAUSED:
            gst_element_query_duration(self.playbin, GST_FORMAT_TIME, &duration)
            return duration

        # preroll
        gst_element_set_state(self.pipeline, GST_STATE_PAUSED)
        gst_element_get_state(self.pipeline, &state, NULL,
                <GstClockTime>GST_SECOND)
        gst_element_query_duration(self.playbin, GST_FORMAT_TIME, &duration)
        gst_element_set_state(self.pipeline, GST_STATE_READY)
        return duration

    cdef gint64 _get_position(self) nogil:
        cdef gint64 position = 0
        if self.playbin == NULL:
            return 0
        if not gst_element_query_position(
                self.playbin, GST_FORMAT_TIME, &position):
            return 0
        return position

    cdef void _seek(self, float percent) nogil:
        cdef GstState current_state, pending_state
        cdef gboolean ret
        cdef gint64 seek_t, duration
        if self.playbin == NULL:
            return

        duration = self._get_duration()
        if duration <= 0:
            seek_t = 0
        else:
            seek_t = <gint64>(percent * duration)
        seek_flags = GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT
        gst_element_get_state(self.pipeline, &current_state,
                &pending_state, <GstClockTime>GST_SECOND)
        if current_state == GST_STATE_READY:
            gst_element_set_state(self.pipeline, GST_STATE_PAUSED)
        ret = gst_element_seek_simple(self.playbin, GST_FORMAT_TIME,
                <GstSeekFlags>seek_flags, seek_t)

        if not ret:
            return

        if self.appsink != NULL:
            gst_element_get_state(self.pipeline, &current_state,
                    &pending_state, <GstClockTime>GST_SECOND)
            if current_state != GST_STATE_PLAYING:
                c_appsink_pull_preroll(
                    self.appsink, _on_appsink_sample, <void *>self)



'''
Event dispatcher
================

All objects that produce events in Kivy implement the :class:`EventDispatcher`
which provides a consistent interface for registering and manipulating event
handlers.

.. versionchanged:: 1.0.9
    Property discovery and methods have been moved from the
    :class:`~kivy.uix.widget.Widget` to the :class:`EventDispatcher`.
'''

__all__ = ('EventDispatcher', 'ObjectWithUid', 'Observable')

cimport cython
from libc.stdlib cimport malloc, free
from libc.string cimport memset

from functools import partial
from collections import defaultdict
from kivy.weakmethod import WeakMethod
from kivy.weakproxy import WeakProxy
from kivy.compat import string_types
from kivy.properties cimport (Property, PropertyStorage, ObjectProperty,
    NumericProperty, StringProperty, ListProperty, DictProperty,
    BooleanProperty)
from kivy.utils import deprecated

cdef int widget_uid = 0
cdef dict cache_properties = {}
cdef dict cache_properties_per_cls = {}
cdef dict cache_events = {}
cdef dict cache_events_handlers = {}


cdef class ObjectWithUid(object):
    '''
    (internal) This class assists in providing unique identifiers for class
    instances. It is not intended for direct usage.
    '''
    def __cinit__(self):
        global widget_uid

        # XXX for the moment, we need to create a unique id for properties.
        # Properties need a identifier to the class instance. hash() and id()
        # are longer than using a custom __uid. I hope we can figure out a way
        # of doing that without using any python code. :)
        widget_uid += 1
        self.uid = widget_uid


cdef class Observable(ObjectWithUid):
    ''':class:`Observable` is a stub class defining the methods required
    for binding. :class:`EventDispatcher` is (the) one example of a class that
    implements the binding interface. See :class:`EventDispatcher` for details.

    .. versionadded:: 1.9.0
    '''

    def __cinit__(self, *largs, **kwargs):
        self.__fbind_mapping = defaultdict(list)
        self.bound_uid = 1

    def bind(self, **kwargs):
        pass

    def unbind(self, **kwargs):
        pass

    def fbind(self, name, func, *largs, **kwargs):
        '''See :meth:`EventDispatcher.fbind`.

        .. note::

            To keep backward compatibility with derived classes which may have
            inherited from :class:`Observable` before, the
            :meth:`fbind` method was added. The default implementation
            of :meth:`fbind` is to create a partial
            function that it passes to bind while saving the uid and largs/kwargs.
            However, :meth:`funbind` (and :meth:`unbind_uid`) are fairly
            inefficient since we have to first lookup this partial function
            using the largs/kwargs or uid and then call :meth:`unbind` on
            the returned function. It is recommended to overwrite
            these methods in derived classes to bind directly for
            better performance.

            Similarly to :meth:`EventDispatcher.fbind`, this method returns
            0 on failure and a positive unique uid on success. This uid can be
            used with :meth:`unbind_uid`.

        '''
        uid = self.bound_uid
        self.bound_uid += 1
        f = partial(func, *largs, **kwargs)
        self.__fbind_mapping[name].append(((func, largs, kwargs), uid, f))
        try:
            self.bind(**{name: f})
            return uid
        except KeyError:
            return 0

    def funbind(self, name, func, *largs, **kwargs):
        '''See :meth:`fbind` and :meth:`EventDispatcher.funbind`.
        '''
        cdef object f = None
        cdef tuple item, val = (func, largs, kwargs)
        cdef list bound = self.__fbind_mapping[name]

        for i, item in enumerate(bound):
            if item[0] == val:
                f = item[2]
                del bound[i]
                break

        if f is not None:
            try:
                self.unbind(**{name: f})
            except KeyError:
                pass

    def unbind_uid(self, name, uid):
        '''See :meth:`fbind` and :meth:`EventDispatcher.unbind_uid`.
        '''
        cdef object f = None
        cdef tuple item
        cdef list bound = self.__fbind_mapping[name]
        if not uid:
            raise ValueError(
                'uid, {}, that evaluates to False is not valid'.format(uid))

        for i, item in enumerate(bound):
            if item[1] == uid:
                f = item[2]
                del bound[i]
                break

        if f is not None:
            try:
                self.unbind(**{name: f})
            except KeyError:
                pass

    @property
    def proxy_ref(self):
        return self


cdef class EventDispatcher(ObjectWithUid):
    '''Generic event dispatcher interface.

    See the module docstring for usage.
    '''

    def __cinit__(self, *largs, **kwargs):
        global cache_properties
        cdef dict attrs_found
        cdef list attrs
        cdef Property attr
        cdef str k

        self.__event_stack = {}
        self.__storage = {}
        self._proxy_ref = None

        __cls__ = self.__class__

        # the props are filled in by Property.__set_name__. If we have properties
        # with same name declared in different sub-classes, the one soonest in
        # the mro wins, so we must compile props in reverse mro order so earliest wins
        if __cls__ not in cache_properties:
            attrs_found = cache_properties[__cls__] = {
                name: prop
                for cls in reversed(__cls__.__mro__)
                for name, prop in cache_properties_per_cls.get(cls, {}).items()
            }
        else:
            attrs_found = cache_properties[__cls__]
        self.__properties = attrs_found

        # now that they have their names, we can link those that need it now
        for k, attr in attrs_found.items():
            attr.link_eagerly(self)

        # Automatic registration of event types (instead of calling
        # self.register_event_type)

        # If not done yet, discover __events__ on all the baseclasses
        cdef dict ce = cache_events
        cdef list events
        cdef str event
        if __cls__ not in ce:
            events = []
            for cls in __cls__.__mro__:
                if not hasattr(cls, '__events__'):
                    continue
                for event in cls.__events__:
                    if event in events:
                        continue

                    if event[:3] != 'on_':
                        raise Exception('{} is not an event name in {}'.format(
                            event, __cls__.__name__))

                    # Ensure that the user has at least declared the default handler
                    if not hasattr(self, event):
                        raise Exception(
                            'Missing default handler <%s> in <%s>' % (
                            event, __cls__.__name__))

                    events.append(event)
            ce[__cls__] = events
        else:
            events = ce[__cls__]

        # then auto register
        for event in events:
            self.__event_stack[event] = EventObservers.__new__(
                EventObservers, 1, 0)

    def __init__(self, **kwargs):
        cdef str func, name, key
        cdef dict properties
        cdef dict prop_args

        # Auto bind on own handler if exist
        properties = self.properties()
        prop_args = {
            k: kwargs.pop(k) for k in list(kwargs.keys()) if k in properties}
        self._kwargs_applied_init = set(prop_args.keys()) if prop_args else set()

        # at this point any kwargs passed may go to object
        # which will raise a TypeError and cause confusion
        try:
           super(EventDispatcher, self).__init__(**kwargs)
        except TypeError as e:
           if kwargs and str(e).startswith("object.__init__() takes"):
               raise TypeError(
                  'Properties {} passed to __init__ may not be existing '
                  'property names. Valid properties are {}'.format(
                      sorted(kwargs.keys()), sorted(properties.keys()))) from e
           else:
               raise

        __cls__ = self.__class__
        if __cls__ not in cache_events_handlers:
            event_handlers = []
            for func in dir(self):
                if func[:3] != 'on_':
                    continue
                name = func[3:]
                if name in properties:
                    event_handlers.append(func)
            cache_events_handlers[__cls__] = event_handlers
        else:
            event_handlers = cache_events_handlers[__cls__]
        for func in event_handlers:
            self.fbind(func[3:], getattr(self, func))

        # Apply the existing arguments to our widget
        for key, value in prop_args.items():
            setattr(self, key, value)

    def register_event_type(self, event_type):
        '''Register an event type with the dispatcher.

        Registering event types allows the dispatcher to validate event handler
        names as they are attached and to search attached objects for suitable
        handlers. Each event type declaration must:

            1. start with the prefix `on_`.
            2. have a default handler in the class.

        Example of creating a custom event::

            class MyWidget(Widget):
                def __init__(self, **kwargs):
                    super(MyWidget, self).__init__(**kwargs)
                    self.register_event_type('on_swipe')

                def on_swipe(self):
                    pass

            def on_swipe_callback(*largs):
                print('my swipe is called', largs)
            w = MyWidget()
            w.dispatch('on_swipe')
        '''

        if event_type[:3] != 'on_':
            raise Exception('A new event must start with "on_"')

        # Ensure that the user has at least declared the default handler
        if not hasattr(self, event_type):
            raise Exception(
                'Missing default handler <%s> in <%s>' % (
                event_type, self.__class__.__name__))

        # Add the event type to the stack
        if event_type not in self.__event_stack:
            self.__event_stack[event_type] = EventObservers.__new__(
                EventObservers, 1, 0)

    @deprecated(msg='Deprecated in 2.1.0, use unregister_event_type instead. '
                    'Will be removed after two releases')
    @cython.binding(True)
    def unregister_event_types(self, event_type):
        self.unregister_event_type(event_type)

    def unregister_event_type(self, event_type):
        '''Unregister an event type in the dispatcher.

        .. versionchanged:: 2.1.0
            Method renamed from `unregister_event_types` to
            `unregister_event_type`.
        '''
        self.__event_stack.pop(event_type, None)

    def is_event_type(self, event_type):
        '''Return True if the event_type is already registered.

        .. versionadded:: 1.0.4
        '''
        return event_type in self.__event_stack

    def bind(self, **kwargs):
        '''Bind an event type or a property to a callback.

        Usage::

            # With properties
            def my_x_callback(obj, value):
                print('on object', obj, 'x changed to', value)
            def my_width_callback(obj, value):
                print('on object', obj, 'width changed to', value)
            self.bind(x=my_x_callback, width=my_width_callback)

            # With event
            def my_press_callback(obj):
                print('event on object', obj)
            self.bind(on_press=my_press_callback)

        In general, property callbacks are called with 2 arguments (the
        object and the property's new value) and event callbacks with
        one argument (the object). The example above illustrates this.

        The following example demonstrates various ways of using the bind
        function in a complete application::

            from kivy.uix.boxlayout import BoxLayout
            from kivy.app import App
            from kivy.uix.button import Button
            from functools import partial


            class DemoBox(BoxLayout):
                """
                This class demonstrates various techniques that can be used for binding to
                events. Although parts could me made more optimal, advanced Python concepts
                are avoided for the sake of readability and clarity.
                """
                def __init__(self, **kwargs):
                    super(DemoBox, self).__init__(**kwargs)
                    self.orientation = "vertical"

                    # We start with binding to a normal event. The only argument
                    # passed to the callback is the object which we have bound to.
                    btn = Button(text="Normal binding to event")
                    btn.bind(on_press=self.on_event)

                    # Next, we bind to a standard property change event. This typically
                    # passes 2 arguments: the object and the value
                    btn2 = Button(text="Normal binding to a property change")
                    btn2.bind(state=self.on_property)

                    # Here we use anonymous functions (a.k.a lambdas) to perform binding.
                    # Their advantage is that you can avoid declaring new functions i.e.
                    # they offer a concise way to "redirect" callbacks.
                    btn3 = Button(text="Using anonymous functions.")
                    btn3.bind(on_press=lambda x: self.on_event(None))

                    # You can also declare a function that accepts a variable number of
                    # positional and keyword arguments and use introspection to determine
                    # what is being passed in. This is very handy for debugging as well
                    # as function re-use. Here, we use standard event binding to a function
                    # that accepts optional positional and keyword arguments.
                    btn4 = Button(text="Use a flexible function")
                    btn4.bind(on_press=self.on_anything)

                    # Lastly, we show how to use partial functions. They are sometimes
                    # difficult to grasp, but provide a very flexible and powerful way to
                    # reuse functions.
                    btn5 = Button(text="Using partial functions. For hardcores.")
                    btn5.bind(on_press=partial(self.on_anything, "1", "2", monthy="python"))

                    for but in [btn, btn2, btn3, btn4, btn5]:
                        self.add_widget(but)

                def on_event(self, obj):
                    print("Typical event from", obj)

                def on_property(self, obj, value):
                    print("Typical property change from", obj, "to", value)

                def on_anything(self, *args, **kwargs):
                    print('The flexible function has *args of', str(args),
                        "and **kwargs of", str(kwargs))


            class DemoApp(App):
                def build(self):
                    return DemoBox()

            if __name__ == "__main__":
                DemoApp().run()

        If a callback has already been bound to a given event or property,
        it won't be added again.

        When binding a method to an event or property, a
        :class:`kivy.weakmethod.WeakMethod` of the callback is saved. That is,
        rather than storing a regular reference, it stores both a weak
        reference to the instance (see Python's :class:`weakref`).

        This has two consequences.

        The first is that the binding will not prevent garbage collection of
        the method's object. The client must maintain a reference to the instance for
        the desired lifetime. The callback reference is silently removed if it
        becomes invalid.

        The second is that when using a decorated method e.g.::

            @my_decorator
            def callback(self, *args):
                pass

        the decorator (``my_decorator`` here) must use `wraps <https://docs.python.org/3/library/functools.html#functools.wraps>`_ internally.
        '''
        cdef EventObservers observers
        cdef Property prop

        for key, value in kwargs.iteritems():
            assert callable(value), '{!r} is not callable'.format(value)
            if key[:3] == 'on_':
                observers = self.__event_stack.get(key)
                if observers is None:
                    continue
                # convert the handler to a weak method
                observers.bind(WeakMethod(value), value, 1)
            else:
                prop = self.__properties[key]
                prop.bind(self, value)

    def unbind(self, **kwargs):
        '''Unbind properties from callback functions with similar usage as
        :meth:`bind`.

        If a callback has been bound to a given event or property multiple
        times, only the first occurrence will be unbound.

        .. note::

            It is safe to use :meth:`unbind` on a function bound with :meth:`fbind`
            as long as that function was originally bound without any keyword and
            positional arguments. Otherwise, the function will fail to be unbound
            and you should use :meth:`funbind` instead.
        '''
        cdef EventObservers observers
        cdef Property prop

        for key, value in kwargs.iteritems():
            if key[:3] == 'on_':
                observers = self.__event_stack.get(key)
                if observers is None:
                    continue
                # it's a ref, and stop on first match
                observers.unbind(value, 1)
            else:
                prop = self.__properties[key]
                prop.unbind(self, value, 1)

    def fbind(self, name, func, *largs, **kwargs):
        '''A method for advanced, and typically faster binding. This method is
        different than :meth:`bind` and is meant for more advanced users and
        internal usage. It can be used as long as the following points are heeded.

        #. As opposed to :meth:`bind`, it does not check that this function and
           largs/kwargs has not been bound before to this name. So binding
           the same callback multiple times will just keep adding it.
        #. Although :meth:`bind` creates a :class:`WeakMethod` of the callback when
           binding to an event or property, this method stores the callback
           directly, unless a keyword argument `ref` with value True is provided
           and then a :class:`WeakMethod` is saved.
           This is useful when there's no risk of a memory leak by storing the
           callback directly.
        #. This method returns a unique positive number if `name` was found and
           bound, and `0`, otherwise. It does not raise an exception, like
           :meth:`bind` if the property `name` is not found. If not zero,
           the uid returned is unique to this `name` and callback and can be
           used with :meth:`unbind_uid` for unbinding.


        When binding a callback with largs and/or kwargs, :meth:`funbind`
        must be used for unbinding. If no largs and kwargs are provided,
        :meth:`unbind` may be used as well. :meth:`unbind_uid` can be used in
        either case.

        This method passes on any caught positional and/or keyword arguments to
        the callback, removing the need to call partial. When calling the
        callback the expended largs are passed on followed by instance/value
        (just instance for kwargs) followed by expended kwargs.

        Following is an example of usage similar to the example in
        :meth:`bind`::

            class DemoBox(BoxLayout):

                def __init__(self, **kwargs):
                    super(DemoBox, self).__init__(**kwargs)
                    self.orientation = "vertical"

                    btn = Button(text="Normal binding to event")
                    btn.fbind('on_press', self.on_event)

                    btn2 = Button(text="Normal binding to a property change")
                    btn2.fbind('state', self.on_property)

                    btn3 = Button(text="A: Using function with args.")
                    btn3.fbind('on_press', self.on_event_with_args, 'right',
                                   tree='birch', food='apple')

                    btn4 = Button(text="Unbind A.")
                    btn4.fbind('on_press', self.unbind_a, btn3)

                    btn5 = Button(text="Use a flexible function")
                    btn5.fbind('on_press', self.on_anything)

                    btn6 = Button(text="B: Using flexible functions with args. For hardcores.")
                    btn6.fbind('on_press', self.on_anything, "1", "2", monthy="python")

                    btn7 = Button(text="Force dispatch B with different params")
                    btn7.fbind('on_press', btn6.dispatch, 'on_press', 6, 7, monthy="other python")

                    for but in [btn, btn2, btn3, btn4, btn5, btn6, btn7]:
                        self.add_widget(but)

                def on_event(self, obj):
                    print("Typical event from", obj)

                def on_event_with_args(self, side, obj, tree=None, food=None):
                    print("Event with args", obj, side, tree, food)

                def on_property(self, obj, value):
                    print("Typical property change from", obj, "to", value)

                def on_anything(self, *args, **kwargs):
                    print('The flexible function has *args of', str(args),
                        "and **kwargs of", str(kwargs))
                    return True

                def unbind_a(self, btn, event):
                    btn.funbind('on_press', self.on_event_with_args, 'right',
                                    tree='birch', food='apple')

        .. note::

            Since the kv lang uses this method to bind, one has to implement
            this method, instead of :meth:`bind` when creating a non
            :class:`EventDispatcher` based class used with the kv lang. See
            :class:`Observable` for an example.

        .. versionadded:: 1.9.0

        .. versionchanged:: 1.9.1
            The `ref` keyword argument has been added.
        '''
        cdef EventObservers observers
        cdef Property prop

        if name[:3] == 'on_':
            observers = self.__event_stack.get(name)
            if observers is not None:
                if kwargs.pop('ref', False):
                    return observers.fbind(WeakMethod(func), largs, kwargs, 1)
                else:
                    return observers.fbind(func, largs, kwargs, 0)
            return 0
        else:
            prop = self.__properties.get(name)
            if prop is None:
                return 0
            return prop.fbind(self, func, kwargs.pop('ref', False), largs, kwargs)

    def funbind(self, name, func, *largs, **kwargs):
        '''Similar to :meth:`fbind`.

        When unbinding, :meth:`unbind` will unbind all callbacks that match the
        callback, while this method will only unbind the first.

        To unbind, the same positional and keyword arguments passed to
        :meth:`fbind` must be passed on to funbind.

        .. note::

            It is safe to use :meth:`funbind` to unbind a function bound with
            :meth:`bind` as long as no keyword and positional arguments are
            provided to :meth:`funbind`.

        .. versionadded:: 1.9.0
        '''
        cdef EventObservers observers
        cdef Property prop

        if name[:3] == 'on_':
            observers = self.__event_stack.get(name)
            if observers is not None:
                observers.funbind(func, largs, kwargs)
        else:
            prop = self.__properties.get(name)
            if prop is not None:
                prop.funbind(self, func, largs, kwargs)

    def unbind_uid(self, name, uid):
        '''Uses the uid returned by :meth:`fbind` to unbind the callback.

        This method is much more efficient than :meth:`funbind`. If `uid`
        evaluates to False (e.g. 0) a `ValueError` is raised. Also, only
        callbacks bound with :meth:`fbind` can be unbound with this method.

        Since each call to :meth:`fbind` will generate a unique `uid`,
        only one callback will be removed. If `uid` is not found among the
        callbacks, no error is raised.

        E.g.::

            btn6 = Button(text="B: Using flexible functions with args. For hardcores.")
            uid = btn6.fbind('on_press', self.on_anything, "1", "2", monthy="python")
            if not uid:
                raise Exception('Binding failed').
            ...
            btn6.unbind_uid('on_press', uid)

        .. versionadded:: 1.9.0
        '''
        cdef EventObservers observers
        cdef Property prop

        if name[:3] == 'on_':
            observers = self.__event_stack.get(name)
            if observers is not None:
                observers.unbind_uid(uid)
        else:
            prop = self.__properties.get(name)
            if prop is not None:
                prop.unbind_uid(self, uid)

    def get_property_observers(self, name, args=False):
        ''' Returns a list of methods that are bound to the property/event
        passed as the *name* argument::

            widget_instance.get_property_observers('on_release')

        :Parameters:

            `name`: str
                The name of the event or property.
            `args`: bool
                Whether to return the bound args. To keep compatibility,
                only the callback functions and not their provided args will
                be returned in the list when `args` is False.

                If True, each element in the list is a 5-tuple of
                `(callback, largs, kwargs, is_ref, uid)`, where `is_ref` indicates
                whether `callback` is a weakref, and `uid` is the uid given by
                :meth:`fbind`, or None if :meth:`bind` was used. Defaults to `False`.

        :Returns:
            The list of bound callbacks. See `args` for details.

        .. versionadded:: 1.8.0

        .. versionchanged:: 1.9.0
            `args` has been added.
        '''
        cdef Property prop
        cdef PropertyStorage ps
        cdef EventObservers observers

        if name[:3] == 'on_':
            observers = self.__event_stack[name]
        else:
            prop = self.__properties[name]
            ps = prop.get_property_storage(self)
            observers = ps.observers
        return list(observers) if args else [item[0] for item in observers]

    def events(EventDispatcher self):
        '''Return all the events in the class. Can be used for introspection.

        .. versionadded:: 1.8.0

        '''
        return self.__event_stack.keys()

    def dispatch(self, event_type, *largs, **kwargs):
        '''Dispatch an event across all the handlers added in bind/fbind().
        As soon as a handler returns True, the dispatching stops.

        The function collects all the positional and keyword arguments and
        passes them on to the handlers.

        .. note::
            The handlers are called in reverse order than they were registered
            with :meth:`bind`.

        :Parameters:
            `event_type`: str
                the event name to dispatch.

        .. versionchanged:: 1.9.0
            Keyword arguments collection and forwarding was added. Before, only
            positional arguments would be collected and forwarded.

        '''
        cdef EventObservers observers = self.__event_stack[event_type]
        if observers.dispatch(self, None, largs, kwargs, 1):
            return True

        handler = getattr(self, event_type)
        return handler(*largs, **kwargs)

    def dispatch_generic(self, event_type, *largs, **kwargs):
        if event_type in self.__event_stack:
            return self.dispatch(event_type, *largs, **kwargs)
        return self.dispatch_children(event_type, *largs, **kwargs)

    def dispatch_children(self, event_type, *largs, **kwargs):
        for child in self.children[:]:
            if child.dispatch_generic(event_type, *largs, **kwargs):
                return True

    #
    # Properties
    #
    def __proxy_setter(self, EventDispatcher dstinstance, name, instance, value):
        cdef Property prop = self.__properties[name]
        prop.set(dstinstance, value)

    def __proxy_getter(self, EventDispatcher dstinstance, name, instance):
        cdef Property prop = self.__properties[name]
        return prop.get(dstinstance)

    def setter(self, name):
        '''Return the setter of a property. Use: instance.setter('name').
        The setter is a convenient callback function useful if you want
        to directly bind one property to another.
        It returns a partial function that will accept
        (obj, value) args and results in the property 'name' of instance
        being set to value.

        .. versionadded:: 1.0.9

        For example, to bind number2 to number1 in python you would do::

            class ExampleWidget(Widget):
                number1 = NumericProperty(None)
                number2 = NumericProperty(None)

                def __init__(self, **kwargs):
                    super(ExampleWidget, self).__init__(**kwargs)
                    self.bind(number1=self.setter('number2'))

        This is equivalent to kv binding::

            <ExampleWidget>:
                number2: self.number1

        '''
        return partial(self.__proxy_setter, self, name)

    def getter(self, name):
        '''Return the getter of a property.

        .. versionadded:: 1.0.9
        '''
        return partial(self.__proxy_getter, self, name)

    def property(self, name, quiet=False):
        '''Get a property instance from the property name. If quiet is True,
        None is returned instead of raising an exception when `name` is not a
        property. Defaults to `False`.

        .. versionadded:: 1.0.9

        :return:

            A :class:`~kivy.properties.Property` derived instance
            corresponding to the name.

        .. versionchanged:: 1.9.0
            quiet was added.
        '''
        if quiet:
            return self.__properties.get(name, None)
        else:
            return self.__properties[name]

    cpdef dict properties(EventDispatcher self):
        '''Return all the properties in the class in a dictionary of
        key/property class. Can be used for introspection.

        .. versionadded:: 1.0.9
        '''
        # fast path, use the cache first
        __cls__ = self.__class__
        if __cls__ in cache_properties:
            return cache_properties[__cls__]

        return dict(self.__properties)

    def create_property(self, str name, value=None, default_value=True, *largs, **kwargs):
        '''Create a new property at runtime.

        .. versionadded:: 1.0.9

        .. versionchanged:: 1.8.0
            `value` parameter added, can be used to set the default value of the
            property. Also, the type of the value is used to specialize the
            created property.

        .. versionchanged:: 1.9.0
            In the past, if `value` was of type `bool`, a `NumericProperty`
            would be created, now a `BooleanProperty` is created.

            Also, now and positional and keyword arguments are passed to the
            property when created.

        .. versionchanged:: 2.0.0

            default_value has been added.

        .. warning::

            This function is designed for the Kivy language, don't use it in
            your code. You should declare the property in your class instead of
            using this method.

        :Parameters:
            `name`: string
                Name of the property
            `value`: object, optional
                Default value of the property. Type is also used for creating
                more appropriate property types. Defaults to None.
            `default_value`: bool, True by default
                If True, `value` will be the default for the property. Otherwise,
                the property will be initialized with the the property type's
                normal default value, and subsequently set to ``value``.

        ::

            >>> mywidget = Widget()
            >>> mywidget.create_property('custom')
            >>> mywidget.custom = True
            >>> print(mywidget.custom)
            True
        '''
        cdef Property prop
        if value is None:  # shortcut
            cls = ObjectProperty
        elif isinstance(value, bool):
            cls = BooleanProperty
        elif isinstance(value, (int, float)):
            cls = NumericProperty
        elif isinstance(value, string_types):
            cls = StringProperty
        elif isinstance(value, (list, tuple)):
            cls = ListProperty
        elif isinstance(value, dict):
            cls = DictProperty
        else:
            cls = ObjectProperty

        if default_value:
            prop = cls(value, *largs, **kwargs)
        else:
            prop = cls(*largs, **kwargs)

        self.__properties[name] = prop
        prop.set_name(self, name)
        prop.link_eagerly(self)
        setattr(self.__class__, name, prop)

        if not default_value:
            setattr(self, name, value)

    def apply_property(self, **kwargs):
        '''Adds properties at runtime to the class. The function accepts
        keyword arguments of the form `prop_name=prop`, where `prop` is a
        :class:`Property` instance and `prop_name` is the name of the attribute
        of the property.

        .. versionadded:: 1.9.1

        .. warning::

            This method is not recommended for common usage because you should
            declare the properties in your class instead of using this method.

        For example::

            >>> print(wid.property('sticks', quiet=True))
            None
            >>> wid.apply_property(sticks=ObjectProperty(55, max=10))
            >>> print(wid.property('sticks', quiet=True))
            <kivy.properties.ObjectProperty object at 0x04303130>
        '''
        cdef Property prop
        cdef str name
        for name, prop in kwargs.items():
            self.__properties[name] = prop
            prop.set_name(self, name)
            prop.link_eagerly(self)
            setattr(self.__class__, name, prop)

    @property
    def proxy_ref(self):
        '''Returns a :class:`~kivy.weakproxy.WeakProxy` reference to the
        :class:`EventDispatcher`.

        .. versionadded:: 1.9.0

        .. versionchanged:: 2.0.0

            Previously it just returned itself, now it actually returns a
            :class:`~kivy.weakproxy.WeakProxy`.
        '''
        _proxy_ref = self._proxy_ref
        if _proxy_ref is not None:
            return _proxy_ref

        self._proxy_ref = _proxy_ref = WeakProxy(self)
        return _proxy_ref

    @property
    def __self__(self):
        return self


cdef class BoundCallback:
    """Bound callback storage.

    .. note::

        We do not call ``__init__`` on the class because we use cython's faster
        instantiation using ``__new__``.
    """

    def __cinit__(
            self, EventObservers observers, object func, tuple largs,
            dict kwargs, int is_ref, uid=None):
        self.observers = observers
        self.func = func
        self.largs = largs
        self.kwargs = kwargs
        self.is_ref = is_ref
        self.lock = unlocked
        self.prev = self.next = None
        self.uid = uid

    def unbind_callback(self, *args):
        self.observers.unbind_callback(self)

    cdef void set_largs(self, tuple largs):
        self.largs = largs


cdef class EventObservers:
    '''A class that stores observers as a doubly linked list. See dispatch
    for more details on locking and deletion of observers.

    In all instances, largs and kwargs if None or empty are all converted
    to None internally before storing or comparing.

    .. note::

        We do not call ``__init__`` on the class because we use cython's faster
        instantiation using ``__new__``.
    '''

    def __cinit__(self, int dispatch_reverse=0, dispatch_value=1):
        self.dispatch_reverse = dispatch_reverse
        self.dispatch_value = dispatch_value
        self.last_callback = self.first_callback = None
        self.uid = 1  # start with 1 so uid is always evaluated to True

    cdef inline BoundCallback make_callback(
            self, object observer, tuple largs, dict kwargs, int is_ref, uid=None):
        return BoundCallback.__new__(
            BoundCallback, self, observer,
            largs if largs else None,
            kwargs if kwargs else None,
            is_ref, uid)

    cdef inline void bind(self, object observer, object src_observer, int is_ref) except *:
        '''Bind the observer to the event. If this observer has already been
        bound, we don't add it again.
        '''
        cdef BoundCallback callback = self.first_callback
        cdef BoundCallback new_callback
        cdef int cb_equal

        while callback is not None:
            if is_ref and not callback.is_ref:
                cb_equal = callback.func == src_observer
            elif callback.is_ref and not is_ref:
                cb_equal = callback.func() == observer
            elif is_ref:
                cb_equal = callback.func() == src_observer
            else:
                cb_equal = callback.func == observer
            if (callback.lock != deleted and callback.largs is None and
                callback.kwargs is None and cb_equal):
                return
            callback = callback.next

        new_callback = self.make_callback(observer, None, None, is_ref)
        self.fbind_existing_callback(new_callback)

    cdef inline object fbind(self, object observer, tuple largs, dict kwargs,
                               int is_ref):
        '''Similar to bind, except it accepts largs, kwargs that is forwards.
        is_ref, if true, will mark the observer that it is a ref so that we
        can unref it before calling.
        '''
        cdef object uid = self.uid
        self.uid += 1
        cdef BoundCallback new_callback = self.make_callback(
            observer, largs, kwargs, is_ref, uid)
        self.fbind_existing_callback(new_callback)
        return uid

    cdef inline BoundCallback fbind_callback(
            self, object observer, tuple largs, dict kwargs, int is_ref):
        '''Similar to bind, except it accepts largs, kwargs that is forwards.
        is_ref, if true, will mark the observer that it is a ref so that we
        can unref it before calling.
        '''
        cdef BoundCallback new_callback = self.make_callback(
            observer, largs, kwargs, is_ref)
        self.fbind_existing_callback(new_callback)
        return new_callback

    cdef inline void fbind_existing_callback(self, BoundCallback callback):
        if self.first_callback is None:
            self.last_callback = self.first_callback = callback
        else:
            self.last_callback.next = callback
            callback.prev = self.last_callback
            self.last_callback = callback

    cdef inline void unbind(self, object observer, int stop_on_first) except *:
        '''Removes the observer. If is_ref, he observers will be derefed before
        comparing to observer, if they are refed. If stop_on_first, after the
        first match we return.
        '''
        cdef object f
        cdef BoundCallback callback = self.first_callback

        while callback is not None:
            # try a quick comparison
            if callback.lock == deleted or callback.largs is not None or callback.kwargs is not None:
                callback = callback.next
                continue

            # now match the actual callback function
            if callback.is_ref:
                f = callback.func()
            else:
                f = callback.func
            if f != observer:
                callback = callback.next
                continue

            self.remove_callback(callback)
            callback = callback.next

            if stop_on_first:
                return

    cdef inline void funbind(self, object observer, tuple largs, dict kwargs) except *:
        '''Similar to unbind, except we only remove the first match, and
        we don't deref the observers before comparing to observer. The
        largs and kwargs must match the largs and kwargs from when binding.
        '''
        cdef BoundCallback callback = self.first_callback
        largs = largs if largs else None
        kwargs = kwargs if kwargs else None

        while callback is not None:
            if (callback.lock == deleted or
                not callback.is_ref and callback.func != observer or
                callback.is_ref and callback.func() != observer or
                callback.largs != largs or callback.kwargs != kwargs):
                callback = callback.next
                continue

            self.remove_callback(callback)
            return

    cdef inline object unbind_uid(self, object uid):
        '''Remove the callback identified by the uid. If passed uid is None,
        a ValueError is raised.
        '''
        cdef BoundCallback callback = self.first_callback
        if not uid:
            raise ValueError(
                'uid, {}, that evaluates to False is not valid'.format(uid))

        while callback is not None:
            if callback.uid != uid:
                callback = callback.next
                continue

            if callback.lock != deleted:
                self.remove_callback(callback)
            return

    cdef inline object unbind_callback(self, BoundCallback callback):
        '''Remove the callback identified by the uid. If passed uid is None,
        a ValueError is raised.
        '''
        if callback.lock != deleted:
            self.remove_callback(callback)

    cdef inline void remove_callback(self, BoundCallback callback, int force=0) except *:
        '''Removes the callback from the doubly linked list. If the callback is
        locked, unless forced, the lock is changed to deleted and the callback
        is not removed.

        Assumes that callback.lock is either locked, or unlocked, not deleted
        except if force, then it can be anything.
        '''
        if callback.lock == locked and not force:
            callback.lock = deleted
        else:
            if callback.prev is not None:
                callback.prev.next = callback.next
            else:
                self.first_callback = callback.next
            if callback.next is not None:
                callback.next.prev = callback.prev
            else:
                self.last_callback = callback.prev

    cdef inline object _dispatch(
        self, object f, tuple slargs, dict skwargs, object obj, object value,
        tuple largs, dict kwargs):
        '''Dispatches the the callback with the args. f is the (derefed)
        callback. slargs, skwargs are the bound-time provided args. largs, kwargs
        are the dispatched args. The order of args is slargs, obj, value,
        skwargs updated with kwargs. If dispatch_value is False, value is skipped.
        '''
        cdef object result
        cdef dict d
        cdef tuple param = (obj, value) if self.dispatch_value else (obj, )
        cdef tuple fargs = None

        if slargs is not None and skwargs is not None:  # both kw and largs
            if largs is not None:
                fargs = slargs + param + largs
            else:
                fargs = slargs + param

            if kwargs is not None:
                d = dict(skwargs)
                d.update(kwargs)
            else:
                d = skwargs

            return f(*fargs, **d)
        elif slargs is not None:  # only largs
            if largs is not None:
                fargs = slargs + param + largs
            else:
                fargs = slargs + param

            if kwargs is None:
                return f(*fargs)
            else:
                return f(*fargs, **kwargs)
        elif skwargs is not None:  # only kwargs
            if kwargs is not None:
                d = dict(skwargs)
                d.update(kwargs)
            else:
                d = skwargs

            if largs is None:
                if self.dispatch_value:
                    return f(obj, value, **d)
                else:
                    return f(obj, **d)
            else:
                if self.dispatch_value:
                    return f(obj, value, *largs, **d)
                else:
                    return f(obj, *largs, **d)
        else:  # no args
            if largs is None:
                if kwargs is None:
                    if self.dispatch_value:
                        return f(obj, value)
                    else:
                        return f(obj)
                else:
                    if self.dispatch_value:
                        return f(obj, value, **kwargs)
                    else:
                        return f(obj, **kwargs)
            else:
                if kwargs is None:
                    if self.dispatch_value:
                        return f(obj, value, *largs)
                    else:
                        return f(obj, *largs)
                else:
                    if self.dispatch_value:
                        return f(obj, value, *largs, **kwargs)
                    else:
                        return f(obj, *largs, **kwargs)

    cdef inline int dispatch(self, object obj, object value, tuple largs,
                             dict kwargs, int stop_on_true) except 2:
        '''Dispatches obj, value to all bound observers. If largs and/or kwargs,
        they are forwarded after obj, value. if stop_on_true, if a observer returns
        true, the function stops and returns true.

        If dispatch_reverse is True, we dispatch starting with last bound callback,
        otherwise we start with the first.

        The logic and reason for locking callbacks is as followes. During a dispatch,
        arbitrary code can be executed, therefore, as we traverse and execute
        each callback, the callback may in turn bind. unbind or even cause a
        new dispatch recursively many times. Therefore, our goal should be to
        during a dispatch, allow such recursiveness, while at each level, only
        dispatch the callbacks that existed when we started dispatching, but
        not including callbacks removed during dispatching.

        Essentially, we want to make a copy of the callbacks as exited during
        start of dispatching, while allowing removal of callbacks. With a python
        list, we'd have to make a copy of the list and before each callback, we
        check the original list to see if the callback has been removed. We solve
        this issue for the doubly linked list using locks.

        At each recursion level, if a callback is already locked by a higher level,
        we can mark it deleted but not actually delete it or unlock it. Also, that level
        is responsible for deleting the callbacks it locked if a lower
        level marked them deleted, otherwise it just unlocks them before returning.
        So a callback locked by a level, is guaranteed to not be removed (but at most
        marked for deletion) by a recursive dispatch.

        Each callback as it is dispatched is locked. Also, the last callback
        scheduled to be executed is immediately locked, so that we know where to
        stop, in case new callbacks are added.
        '''
        cdef BoundCallback callback, final
        cdef object f, result
        cdef BoundLock current_lock, last_lock
        cdef int done = 0, res = 0, reverse = self.dispatch_reverse

        if reverse:  # dispatch starting from last until first
            callback = self.last_callback  # start callback
            final = self.first_callback  # last callback
        else:
            callback = self.first_callback
            final = self.last_callback
        if callback is None:
            return 0


        last_lock = final.lock  # save the state of the lock of final callback
        if last_lock == unlocked:  # lock the final callback
            final.lock = locked

        while not done and callback is not None:
            done = final is callback

            if callback.lock == deleted:
                callback = callback.prev if reverse else callback.next
                continue

            # save the lock state (currently only either locked or unlocked)
            current_lock = callback.lock
            if current_lock == unlocked:  # and lock it if unlocked
                callback.lock = locked

            if callback.is_ref:
                f = callback.func()
                if f is None:
                    self.remove_callback(callback, current_lock == unlocked)
                    callback = callback.prev if reverse else callback.next
                    continue
            else:
                f = callback.func

            result = self._dispatch(
                f, callback.largs, callback.kwargs, obj, value, largs, kwargs)

            if current_lock == unlocked:  # now unlock/delete if it was unlocked
                if callback.lock == deleted:
                    self.remove_callback(callback, 1)
                else:
                    callback.lock = unlocked

            if result and stop_on_true:
                res = done = 1
            callback = callback.prev if reverse else callback.next

        # now unlock/delete the final callback if we locked it
        if last_lock == unlocked:
            if final.lock == deleted:
                self.remove_callback(final, 1)
            else:
                final.lock = unlocked
        return res

    def __iter__(self):
        '''Binding/unbinding/dispatching while iterating can lead to invalid
        data.
        '''
        cdef BoundCallback callback = self.first_callback

        while callback is not None:
            yield (
                callback.func, callback.largs if callback.largs is not None else (),
                callback.kwargs if callback.kwargs is not None else {},
                callback.is_ref, callback.uid)
            callback = callback.next



#cython: c_string_type=unicode, c_string_encoding=utf8
'''
OpenGL utilities
================

.. versionadded:: 1.0.7
'''

__all__ = ('gl_get_extensions', 'gl_has_extension',
        'gl_has_capability', 'gl_register_get_size',
        'gl_has_texture_format', 'gl_has_texture_conversion',
        'gl_has_texture_native_format', 'gl_get_texture_formats',
        'gl_get_version', 'gl_get_version_minor', 'gl_get_version_major',
        'GLCAP_BGRA', 'GLCAP_NPOT', 'GLCAP_S3TC', 'GLCAP_DXT1', 'GLCAP_ETC1')

include "../include/config.pxi"
include "opengl_utils_def.pxi"

from kivy.graphics.cgl cimport *
from kivy.logger import Logger
from kivy.utils import platform
from kivy.graphics.opengl import _GL_GET_SIZE


cdef list _gl_extensions = []
cdef dict _gl_caps = {}
cdef tuple _gl_texture_fmts = (
    'rgb', 'rgba', 'luminance', 'luminance_alpha',
    'bgr', 'bgra', 's3tc_dxt1', 's3tc_dxt3', 's3tc_dxt5',
    'pvrtc_rgb4', 'pvrtc_rgb2', 'pvrtc_rgba4', 'pvrtc_rgba2')
cdef int _gl_version_major = -1
cdef int _gl_version_minor = -1
cdef str _platform = str(platform)


cpdef list gl_get_extensions():
    '''Return a list of OpenGL extensions available. All the names in the list
    have the `GL_` stripped at the start (if it exists) and are in lowercase.

    >>> print(gl_get_extensions())
    ['arb_blend_func_extended', 'arb_color_buffer_float', 'arb_compatibility',
     'arb_copy_buffer'... ]

    '''
    global _gl_extensions
    cdef str extensions
    if not _gl_extensions:
        extensions = <char *>cgl.glGetString(GL_EXTENSIONS)
        _gl_extensions[:] = [x[3:].lower() if x[:3] == 'GL_' else x.lower()\
                for x in extensions.split()]
    return _gl_extensions


cpdef int gl_has_extension(name):
    '''Check if an OpenGL extension is available. If the name starts with `GL_`,
    it will be stripped for the test and converted to lowercase.

        >>> gl_has_extension('NV_get_tex_image')
        False
        >>> gl_has_extension('OES_texture_npot')
        True

    '''
    if cgl_get_initialized_backend_name() == "mock":
        return True
    name = name.lower()
    if name.startswith('GL_'):
        name = name[3:]
    return name in gl_get_extensions()


cpdef gl_register_get_size(int constid, int size):
    '''Register an association between an OpenGL Const used in glGet* to a number
    of elements.

    By example, the GPU_MEMORY_INFO_DEDICATED_VIDMEM_NVX is a special pname that
    will return the integer 1 (nvidia only).

        >>> GPU_MEMORY_INFO_DEDICATED_VIDMEM_NVX = 0x9047
        >>> gl_register_get_size(GPU_MEMORY_INFO_DEDICATED_VIDMEM_NVX, 1)
        >>> glGetIntegerv(GPU_MEMORY_INFO_DEDICATED_VIDMEM_NVX)[0]
        524288

    '''
    _GL_GET_SIZE[constid] = size


cpdef int gl_has_capability(int cap):
    '''Return the status of a OpenGL Capability. This is a wrapper that
    auto-discovers all the capabilities that Kivy might need. The current
    capabilities tested are:

        - GLCAP_BGRA: Test the support of BGRA texture format
        - GLCAP_NPOT: Test the support of Non Power of Two texture
        - GLCAP_S3TC: Test the support of S3TC texture (DXT1, DXT3, DXT5)
        - GLCAP_DXT1: Test the support of DXT texture (subset of S3TC)
        - GLCAP_ETC1: Test the support of ETC1 texture

    '''
    cdef int value = _gl_caps.get(cap, -1)
    cdef str msg, sval

    # if we got a value, it's already initialized, return it!
    if value != -1:
        return value

    # ok, never been initialized, do it now.
    if cap == c_GLCAP_BGRA:
        msg = 'BGRA texture support'
        if _platform == 'ios':
            value = gl_has_extension('APPLE_texture_format_BGRA8888')
        else:
            value = gl_has_extension('EXT_bgra')
        if not value:
            value = gl_has_extension('EXT_texture_format_BGRA888')

    elif cap == c_GLCAP_NPOT:
        msg = 'NPOT texture support'
        if _platform == 'ios' or _platform == 'android':
            # Adreno 200 renderer doesn't support NPOT
            sval = <char *>cgl.glGetString(GL_RENDERER)
            if sval == 'Adreno 200':
                value = 0
            else:
                value = 1
        else:
            value = gl_has_extension('ARB_texture_non_power_of_two')
            if not value:
                value = gl_has_extension('OES_texture_npot')
            if not value:
                # motorola droid don't have OES_ but IMG_
                value = gl_has_extension('IMG_texture_npot')

    elif cap == c_GLCAP_S3TC:
        # S3TC support DXT1, DXT3 and DXT5
        msg = 'S3TC texture support'
        value = gl_has_extension('S3_s3tc')
        if not value:
            value = gl_has_extension('EXT_texture_compression_s3tc')
        if not value:
            value = gl_has_extension('OES_texture_compression_s3tc')

    elif cap == c_GLCAP_DXT1:
        # DXT1 is included inside S3TC, but not the inverse.
        msg = 'DXT1 texture support'
        value = gl_has_capability(c_GLCAP_S3TC)
        if not value:
            value = gl_has_extension('EXT_texture_compression_dxt1')

    elif cap == c_GLCAP_PVRTC:
        # PVRTC = PowerVR, mostly available in iOS device
        msg = 'PVRTC texture support'
        value = gl_has_extension('IMG_texture_compression_pvrtc')

    elif cap == c_GLCAP_ETC1:
        # PVRTC = PowerVR, mostly available in iOS device
        msg = 'ETC1 texture support'
        value = gl_has_extension('OES_compressed_ETC1_RGB8_texture')

    elif cap == c_GLCAP_UNPACK_SUBIMAGE:
        # Is GL_UNPACK_ROW_LENGTH is supported
        msg = 'Unpack subimage support'
        if _platform == 'ios' or _platform == 'android':
            value = gl_has_extension('EXT_unpack_subimage')
        else:
            value = 1

    else:
        raise Exception('Unknown capability')

    _gl_caps[cap] = value
    if value:
        Logger.info('GL: %s is available' % msg)
    else:
        Logger.warning('GL: %s is not available' % msg)

    return value


cpdef tuple gl_get_texture_formats():
    '''Return a list of texture formats recognized by kivy.
    The texture list is informative but might not been supported by your
    hardware. If you want a list of supported textures, you must filter that
    list as follows::

        supported_fmts = [gl_has_texture_format(x) for x in gl_get_texture_formats()]

    '''
    return _gl_texture_fmts


cpdef int gl_has_texture_native_format(fmt):
    '''Return 1 if the texture format is handled natively.

    >>> gl_has_texture_format('azdmok')
    0
    >>> gl_has_texture_format('rgba')
    1
    >>> gl_has_texture_format('s3tc_dxt1')
    [INFO   ] [GL          ] S3TC texture support is available
    [INFO   ] [GL          ] DXT1 texture support is available
    1

    '''
    if fmt in ('rgb', 'rgba', 'luminance', 'luminance_alpha', 'red', 'rg'):
        return 1
    if fmt in ('palette4_rgb8', 'palette4_rgba8', 'palette4_r5_g6_b5', 'palette4_rgba4', 'palette4_rgb5_a1', 'palette8_rgb8', 'palette8_rgba8', 'palette8_r5_g6_b5', 'palette8_rgba4', 'palette8_rgb5_a1'):
        return gl_has_extension('OES_compressed_paletted_texture')
    if fmt in ('bgr', 'bgra'):
        return gl_has_capability(c_GLCAP_BGRA)
    if fmt == 's3tc_dxt1':
        return gl_has_capability(c_GLCAP_DXT1)
    if fmt.startswith('s3tc_dxt'):
        return gl_has_capability(c_GLCAP_S3TC)
    if fmt.startswith('pvrtc_'):
        return gl_has_capability(c_GLCAP_PVRTC)
    if fmt.startswith('etc1_'):
        return gl_has_capability(c_GLCAP_ETC1)
    return 0


cpdef int gl_has_texture_conversion(fmt):
    '''Return 1 if the texture can be converted to a native format.
    '''
    return fmt in ('bgr', 'bgra', 'argb', 'abgr')


cpdef int gl_has_texture_format(fmt):
    '''Return whether a texture format is supported by your system, natively or
    by conversion. For example, if your card doesn't support 'bgra', we are able
    to convert to 'rgba' but only in software mode.
    '''
    # check if the support of a format can be done natively
    if gl_has_texture_native_format(fmt):
        return 1
    # otherwise, check if it can be converted
    return gl_has_texture_conversion(fmt)


cpdef tuple gl_get_version():
    '''Return the (major, minor) OpenGL version, parsed from the GL_VERSION.

    .. versionadded:: 1.2.0
    '''

    global _gl_version_minor, _gl_version_major
    cdef str version

    if _gl_version_major == -1:

        _gl_version_minor = _gl_version_major = 0
        version = <char *>cgl.glGetString(GL_VERSION)

        try:
            # same parsing algo as Panda3D
            sver = ''
            found = 0
            for c in version:
                if found and c == ' ':
                    break
                if 49 <= ord(c) <= 57:
                    found = 1
                if found:
                    sver += c

            component = sver.split('.')
            if len(component) >= 1:
                _gl_version_major = int(component[0])
            if len(component) >= 2:
                _gl_version_minor = int(component[1])

        except:
            Logger.warning('OpenGL: Error while parsing GL_VERSION')

    return _gl_version_major, _gl_version_minor


cpdef int gl_get_version_major():
    '''Return the major component of the OpenGL version.

    .. versionadded:: 1.2.0
    '''
    if _gl_version_major == -1:
        gl_get_version()
    return _gl_version_major


cpdef int gl_get_version_minor():
    '''Return the minor component of the OpenGL version.

    .. versionadded:: 1.2.0
    '''
    if _gl_version_major == -1:
        gl_get_version()
    return _gl_version_minor



include "common.pxi"

cdef class Buffer:
    '''
    The Buffer class is designed to manage a very fast list of fixed size
    blocks. You can easily add and remove data from the buffer.
    '''
    def __cinit__(self):
        self.data = NULL
        self.i_free = 0
        self.block_size = 0
        self.block_count = 0
        self.l_free = NULL

    def __dealloc__(self):
        if self.data != NULL:
            free(self.data)
            self.data = NULL
        self.block_count = 0
        self.block_size = 0
        if self.l_free != NULL:
            free(self.l_free)

    def __init__(self, long block_size):
        self.block_size = block_size

    cdef void grow(self, long block_count):
        '''Automatically realloc the memory if there are not enough blocks.
        Works only for "grow" operations, not the inverse.
        '''
        cdef void *newptr = NULL
        cdef long i
        cdef int l_free_tmp
        cdef long diff

        # set block_count to the nearest 8 block
        diff = block_count % 8
        if diff != 0:
            block_count = (8 - (block_count % 8)) + block_count

        if block_count <= self.block_count:
            return

        # Try to realloc
        newptr = realloc(self.data, self.block_size * block_count)
        if newptr == NULL:
            raise SystemError('Unable to realloc memory for buffer')

        # Realloc work, put the new pointer
        self.data = newptr

        # Create the free blocks
        newptr = realloc(self.l_free, sizeof(int) * block_count)
        if newptr == NULL:
            raise SystemError('Unable to realloc memory for free list')
        self.l_free = <int *>newptr

        # Initialize the list with index of free block
        for i in xrange(self.block_count, block_count):
            self.l_free[i] = <int>i

        # Update how many block are allocated
        self.block_count = block_count

    cdef void clear(self):
        '''Clear the whole buffer and mark all blocks as available.
        '''
        cdef long i
        for i in xrange(self.block_count):
            self.l_free[i] = <int>i
        self.i_free = 0

    cdef void add(self, void *blocks, unsigned short *indices, int count):
        '''Add a list of blocks inside the buffer.
        '''
        cdef int i, block
        cdef void *p

        # Ensure that our buffer is enough for having all the elements
        if count > self.block_count - self.i_free:
            self.grow(self.block_count + count)

        # Add all the block inside our buffer
        for i in xrange(count):
            p = <void *>(<char *>blocks + (self.block_size * i))

            # Take a free block
            block = self.l_free[self.i_free]
            self.i_free += 1

            # Copy content
            memcpy(<char *>(self.data) + (block * self.block_size), p, self.block_size)

            # Push the current block as indices
            if indices != NULL:
                indices[i] = block

    cdef void remove(self, unsigned short *indices, int count):
        '''Remove a block from the list.
        '''
        cdef int i

        for i in xrange(count):
            # Append the new indice as free block
            self.i_free -= 1
            self.l_free[self.i_free] = indices[i]

    cdef void update(self, int index, void* blocks, int count):
        '''Update *count* number of blocks starting at *index* with the data in
        blocks.
        '''
        memcpy(<char *>(self.data) + (index * self.block_size), blocks, self.block_size * count)

    cdef int count(self):
        '''Return the number of blocks currently used.
        '''
        return self.i_free

    cdef long size(self):
        '''Return the size of the allocated buffer.
        '''
        return self.block_size * self.block_count

    cdef void *pointer(self):
        '''Return the data pointer.
        '''
        return self.data

    cdef void *offset_pointer(self, int offset):
        return <char *>(self.data) + (offset * self.block_size)




'''
Script to generate Kivy API from source code.

Code is messy, but working.
Be careful if you change anything in !

'''

ignore_list = (
    'kivy._clock',
    'kivy._event',
    'kivy.factory_registers',
    'kivy.graphics.buffer',
    'kivy.graphics.vbo',
    'kivy.graphics.vertex',
    'kivy.uix.recycleview.__init__',
    'kivy.setupconfig',
    'kivy.version'
)

import os
import sys
from glob import glob

import kivy

# force loading of kivy modules
import kivy.app
import kivy.metrics
import kivy.atlas
import kivy.context
import kivy.core.audio
import kivy.core.camera
import kivy.core.clipboard
import kivy.core.gl
import kivy.core.image
import kivy.core.spelling
import kivy.core.text
import kivy.core.text.markup
import kivy.core.video
import kivy.core.window
import kivy.geometry
import kivy.graphics
import kivy.graphics.shader
import kivy.graphics.tesselator
import kivy.animation
import kivy.modules.console
import kivy.modules.keybinding
import kivy.modules.monitor
import kivy.modules.touchring
import kivy.modules.inspector
import kivy.modules.recorder
import kivy.modules.screen
import kivy.modules.joycursor
import kivy.storage
import kivy.storage.dictstore
import kivy.storage.jsonstore
import kivy.storage.redisstore
import kivy.network.urlrequest
import kivy.modules.webdebugger
import kivy.support
try:
    import kivy.tools.packaging.pyinstaller_hooks
except ImportError:
    pass
import kivy.input.recorder
import kivy.interactive
import kivy.garden
from kivy.factory import Factory
from kivy.lib import ddsfile, mtdev

# check for silenced build
BE_QUIET = True
if os.environ.get('BE_QUIET') == 'False':
    BE_QUIET = False

# force loading of all classes from factory
for x in list(Factory.classes.keys())[:]:
    getattr(Factory, x)

# Directory of doc
base_dir = os.path.dirname(__file__)
dest_dir = os.path.join(base_dir, 'sources')
examples_framework_dir = os.path.join(base_dir, '..', 'examples', 'framework')

# Check touch file
base = 'autobuild.py-done'
with open(os.path.join(base_dir, base), 'w') as f:
    f.write('')


def writefile(filename, data):
    global dest_dir
    # avoid to rewrite the file if the content didn't change
    f = os.path.join(dest_dir, filename)
    if not BE_QUIET:
        print('write', filename)
    if os.path.exists(f):
        with open(f) as fd:
            if fd.read() == data:
                return
    h = open(f, 'w')
    h.write(data)
    h.close()


# Activate Kivy modules
'''
for k in kivy.kivy_modules.list().keys():
    kivy.kivy_modules.import_module(k)
'''


# Search all kivy module
l = [(x, sys.modules[x],
      os.path.basename(sys.modules[x].__file__).rsplit('.', 1)[0])
      for x in sys.modules if x.startswith('kivy') and sys.modules[x]]


# Extract packages from modules
packages = []
modules = {}
api_modules = []
for name, module, filename in l:
    if name in ignore_list:
        continue
    if not any([name.startswith(x) for x in ignore_list]):
        api_modules.append(name)
    if filename == '__init__':
        packages.append(name)
    else:
        if hasattr(module, '__all__'):
            modules[name] = module.__all__
        else:
            modules[name] = [x for x in dir(module) if not x.startswith('__')]

packages.sort()

# Create index
api_index = '''API Reference
-------------

The API reference is a lexicographic list of all the different classes,
methods and features that Kivy offers.

.. toctree::
    :maxdepth: 1

'''
api_modules.sort()
for package in api_modules:
    api_index += "    api-%s.rst\n" % package

writefile('api-index.rst', api_index)


# Create index for all packages
# Note on displaying inherited members;
#     Adding the directive ':inherited-members:' to automodule achieves this
#     but is not always desired. Please see
#         https://github.com/kivy/kivy/pull/3870

template = '\n'.join((
    '=' * 100,
    '$SUMMARY',
    '=' * 100,
    '''
$EXAMPLES_REF

.. automodule:: $PACKAGE
    :members:
    :show-inheritance:

.. toctree::

$EXAMPLES
'''))


template_examples = '''.. _example-reference%d:

Examples
--------

%s
'''

template_examples_ref = ('# :ref:`Jump directly to Examples'
                         ' <example-reference%d>`')


def extract_summary_line(doc):
    """
    :param doc: the __doc__ field of a module
    :return: a doc string suitable for a header or empty string
    """
    if doc is None:
        return ''
    for line in doc.split('\n'):
        line = line.strip()
        # don't take empty line
        if len(line) < 1:
            continue
        # ref mark
        if line.startswith('.. _'):
            continue
        return line

for package in packages:
    summary = extract_summary_line(sys.modules[package].__doc__)
    if summary is None or summary == '':
        summary = 'NO DOCUMENTATION (package %s)' % package
    t = template.replace('$SUMMARY', summary)
    t = t.replace('$PACKAGE', package)
    t = t.replace('$EXAMPLES_REF', '')
    t = t.replace('$EXAMPLES', '')

    # search packages
    for subpackage in packages:
        packagemodule = subpackage.rsplit('.', 1)[0]
        if packagemodule != package or len(subpackage.split('.')) <= 2:
            continue
        t += "    api-%s.rst\n" % subpackage

    # search modules
    m = list(modules.keys())
    m.sort(key=lambda x: extract_summary_line(sys.modules[x].__doc__).upper())
    for module in m:
        packagemodule = module.rsplit('.', 1)[0]
        if packagemodule != package:
            continue
        t += "    api-%s.rst\n" % module

    writefile('api-%s.rst' % package, t)


# Create index for all module
m = list(modules.keys())
m.sort()
refid = 0
for module in m:
    summary = extract_summary_line(sys.modules[module].__doc__)
    if summary is None or summary == '':
        summary = 'NO DOCUMENTATION (module %s)' % package

    # search examples
    example_output = []
    example_prefix = module
    if module.startswith('kivy.'):
        example_prefix = module[5:]
    example_prefix = example_prefix.replace('.', '_')

    # try to found any example in framework directory
    list_examples = glob('%s*.py' % os.path.join(
        examples_framework_dir, example_prefix))
    for x in list_examples:
        # extract filename without directory
        xb = os.path.basename(x)

        # add a section !
        example_output.append('File :download:`%s <%s>` ::' % (
            xb, os.path.join('..', x)))

        # put the file in
        with open(x, 'r') as fd:
            d = fd.read().strip()
            d = '\t' + '\n\t'.join(d.split('\n'))
            example_output.append(d)

    t = template.replace('$SUMMARY', summary)
    t = t.replace('$PACKAGE', module)
    if len(example_output):
        refid += 1
        example_output = template_examples % (
                refid, '\n\n\n'.join(example_output))
        t = t.replace('$EXAMPLES_REF', template_examples_ref % refid)
        t = t.replace('$EXAMPLES', example_output)
    else:
        t = t.replace('$EXAMPLES_REF', '')
        t = t.replace('$EXAMPLES', '')
    writefile('api-%s.rst' % module, t)


# Generation finished
print('Auto-generation finished')



#
# Kivy - Cross-platform UI framework
# https://kivy.org/
#

import sys
build_examples = False
if "--build_examples" in sys.argv:
    build_examples = True
    sys.argv.remove("--build_examples")

from kivy.utils import pi_version
from copy import deepcopy
import os
from os.path import join, dirname, sep, exists, basename, isdir
from os import walk, environ, makedirs
from distutils.command.build_ext import build_ext
from distutils.version import LooseVersion
from distutils.sysconfig import get_python_inc
from collections import OrderedDict
from time import sleep
from sysconfig import get_paths
from pathlib import Path
import logging
from setuptools import setup, Extension, find_packages


if sys.version_info[0] == 2:
    logging.critical(
        'Unsupported Python version detected!: Kivy 2.0.0 and higher does not '
        'support Python 2. Please upgrade to Python 3, or downgrade Kivy to '
        '1.11.1 - the last Kivy release that still supports Python 2.')


def ver_equal(self, other):
    return self.version == other


# fix error with py3's LooseVersion comparisons
LooseVersion.__eq__ = ver_equal


def get_description():
    with open(join(dirname(__file__), 'README.md'), 'rb') as fileh:
        return fileh.read().decode("utf8").replace('\r\n', '\n')


def getoutput(cmd, env=None):
    import subprocess
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, env=env)
    p.wait()
    if p.returncode:  # if not returncode == 0
        print('WARNING: A problem occurred while running {0} (code {1})\n'
              .format(cmd, p.returncode))
        stderr_content = p.stderr.read()
        if stderr_content:
            print('{0}\n'.format(stderr_content))
        return ""
    return p.stdout.read()


def pkgconfig(*packages, **kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}
    lenviron = None
    pconfig = join(sys.prefix, 'libs', 'pkgconfig')

    if isdir(pconfig):
        lenviron = environ.copy()
        lenviron['PKG_CONFIG_PATH'] = '{};{}'.format(
            environ.get('PKG_CONFIG_PATH', ''), pconfig)
    cmd = 'pkg-config --libs --cflags {}'.format(' '.join(packages))
    results = getoutput(cmd, lenviron).split()
    for token in results:
        ext = token[:2].decode('utf-8')
        flag = flag_map.get(ext)
        if not flag:
            continue
        kw.setdefault(flag, []).append(token[2:].decode('utf-8'))
    return kw


def get_isolated_env_paths():
    try:
        # sdl2_dev is installed before setup.py is run, when installing from
        # source due to pyproject.toml. However, it is installed to a
        # pip isolated env, which we need to add to compiler
        import kivy_deps.sdl2_dev as sdl2_dev
    except ImportError:
        return [], []

    root = os.path.abspath(join(sdl2_dev.__path__[0], '../../../..'))
    includes = [join(root, 'Include')] if isdir(join(root, 'Include')) else []
    libs = [join(root, 'libs')] if isdir(join(root, 'libs')) else []
    return includes, libs


# -----------------------------------------------------------------------------
# Determine on which platform we are

build_examples = build_examples or \
    os.environ.get('KIVY_BUILD_EXAMPLES', '0') == '1'

platform = sys.platform

# Detect Python for android project (http://github.com/kivy/python-for-android)
ndkplatform = environ.get('NDKPLATFORM')
if ndkplatform is not None and environ.get('LIBLINK'):
    platform = 'android'
kivy_ios_root = environ.get('KIVYIOSROOT', None)
if kivy_ios_root is not None:
    platform = 'ios'
# proprietary broadcom video core drivers
if exists('/opt/vc/include/bcm_host.h'):
    used_pi_version = pi_version
    # Force detected Raspberry Pi version for cross-builds, if needed
    if 'KIVY_RPI_VERSION' in environ:
        used_pi_version = int(environ['KIVY_RPI_VERSION'])
    # The proprietary broadcom video core drivers are not available on the
    # Raspberry Pi 4
    if (used_pi_version or 4) < 4:
        platform = 'rpi'
# use mesa video core drivers
if environ.get('VIDEOCOREMESA', None) == '1':
    platform = 'vc'
mali_paths = (
    '/usr/lib/arm-linux-gnueabihf/libMali.so',
    '/usr/lib/arm-linux-gnueabihf/mali-egl/libmali.so',
    '/usr/local/mali-egl/libmali.so')
if any((exists(path) for path in mali_paths)):
    platform = 'mali'

# Needed when cross-compiling
if environ.get('KIVY_CROSS_PLATFORM'):
    platform = environ.get('KIVY_CROSS_PLATFORM')

# -----------------------------------------------------------------------------
# Detect options
#
c_options = OrderedDict()
c_options['use_rpi'] = platform == 'rpi'
c_options['use_egl'] = False
c_options['use_opengl_es2'] = None
c_options['use_opengl_mock'] = environ.get('READTHEDOCS', None) == 'True'
c_options['use_sdl2'] = None
c_options['use_pangoft2'] = None
c_options['use_ios'] = False
c_options['use_android'] = False
c_options['use_mesagl'] = False
c_options['use_x11'] = False
c_options['use_wayland'] = False
c_options['use_gstreamer'] = None
c_options['use_avfoundation'] = platform in ['darwin', 'ios']
c_options['use_osx_frameworks'] = platform == 'darwin'
c_options['debug_gl'] = False

# Set the alpha size, this will be 0 on the Raspberry Pi and 8 on all other
# platforms, so SDL2 works without X11
c_options['kivy_sdl_gl_alpha_size'] = 8 if pi_version is None else 0

# now check if environ is changing the default values
for key in list(c_options.keys()):
    ukey = key.upper()
    if ukey in environ:
        # kivy_sdl_gl_alpha_size should be an integer, the rest are booleans
        value = int(environ[ukey])
        if key != 'kivy_sdl_gl_alpha_size':
            value = bool(value)
        print('Environ change {0} -> {1}'.format(key, value))
        c_options[key] = value

use_embed_signature = environ.get('USE_EMBEDSIGNATURE', '0') == '1'
use_embed_signature = use_embed_signature or bool(
    platform not in ('ios', 'android'))

# -----------------------------------------------------------------------------
# We want to be able to install kivy as a wheel without a dependency
# on cython, but we also want to use cython where possible as a setup
# time dependency through `pyproject.toml` if building from source.

# There are issues with using cython at all on some platforms;
# exclude them from using or declaring cython.

# This determines whether Cython specific functionality may be used.
can_use_cython = True

if platform in ('ios', 'android'):
    # NEVER use or declare cython on these platforms
    print('Not using cython on %s' % platform)
    can_use_cython = False


# -----------------------------------------------------------------------------
# Setup classes

# the build path where kivy is being compiled
src_path = build_path = dirname(__file__)
print("Current directory is: {}".format(os.getcwd()))
print("Source and initial build directory is: {}".format(src_path))

# __version__ is imported by exec, but help linter not complain
__version__ = None
with open(join(src_path, 'kivy', '_version.py'), encoding="utf-8") as f:
    exec(f.read())


class KivyBuildExt(build_ext, object):

    def __new__(cls, *a, **kw):
        # Note how this class is declared as a subclass of distutils
        # build_ext as the Cython version may not be available in the
        # environment it is initially started in. However, if Cython
        # can be used, setuptools will bring Cython into the environment
        # thus its version of build_ext will become available.
        # The reason why this is done as a __new__ rather than through a
        # factory function is because there are distutils functions that check
        # the values provided by cmdclass with issublcass, and so it would
        # result in an exception.
        # The following essentially supply a dynamically generated subclass
        # that mix in the cython version of build_ext so that the
        # functionality provided will also be executed.
        if can_use_cython:
            from Cython.Distutils import build_ext as cython_build_ext
            build_ext_cls = type(
                'KivyBuildExt', (KivyBuildExt, cython_build_ext), {})
            return super(KivyBuildExt, cls).__new__(build_ext_cls)
        else:
            return super(KivyBuildExt, cls).__new__(cls)

    def finalize_options(self):
        retval = super(KivyBuildExt, self).finalize_options()

        # Build the extensions in parallel if the options has not been set
        if hasattr(self, 'parallel') and self.parallel is None:
            # Use a maximum of 4 cores. If cpu_count returns None, then parallel
            # build will be disabled
            self.parallel = min(4, os.cpu_count() or 0)
            if self.parallel:
                print('Building extensions in parallel using {} cores'.format(
                    self.parallel))

        global build_path
        if (self.build_lib is not None and exists(self.build_lib) and
                not self.inplace):
            build_path = self.build_lib
            print("Updated build directory to: {}".format(build_path))

        return retval

    def build_extensions(self):
        # build files
        config_h_fn = ('include', 'config.h')
        config_pxi_fn = ('include', 'config.pxi')
        config_py_fn = ('setupconfig.py', )

        # generate headers
        config_h = '// Autogenerated file for Kivy C configuration\n'
        config_h += '#define __PY3 1\n'
        config_pxi = '# Autogenerated file for Kivy Cython configuration\n'
        config_pxi += 'DEF PY3 = 1\n'
        config_py = '# Autogenerated file for Kivy configuration\n'
        config_py += 'PY3 = 1\n'
        config_py += 'CYTHON_MIN = {0}\nCYTHON_MAX = {1}\n'.format(
            repr(MIN_CYTHON_STRING), repr(MAX_CYTHON_STRING))
        config_py += 'CYTHON_BAD = {0}\n'.format(repr(', '.join(map(
            str, CYTHON_UNSUPPORTED))))

        # generate content
        print('Build configuration is:')
        for opt, value in c_options.items():
            # kivy_sdl_gl_alpha_size is already an integer
            if opt != 'kivy_sdl_gl_alpha_size':
                value = int(bool(value))
            print(' * {0} = {1}'.format(opt, value))
            opt = opt.upper()
            config_h += '#define __{0} {1}\n'.format(opt, value)
            config_pxi += 'DEF {0} = {1}\n'.format(opt, value)
            config_py += '{0} = {1}\n'.format(opt, value)
        debug = bool(self.debug)
        print(' * debug = {0}'.format(debug))

        config_pxi += 'DEF DEBUG = {0}\n'.format(debug)
        config_py += 'DEBUG = {0}\n'.format(debug)
        config_pxi += 'DEF PLATFORM = "{0}"\n'.format(platform)
        config_py += 'PLATFORM = "{0}"\n'.format(platform)
        for fn, content in (
                (config_h_fn, config_h), (config_pxi_fn, config_pxi),
                (config_py_fn, config_py)):
            build_fn = expand(build_path, *fn)
            if self.update_if_changed(build_fn, content):
                print('Updated {}'.format(build_fn))
            src_fn = expand(src_path, *fn)
            if src_fn != build_fn and self.update_if_changed(src_fn, content):
                print('Updated {}'.format(src_fn))

        c = self.compiler.compiler_type
        print('Detected compiler is {}'.format(c))
        if c != 'msvc':
            for e in self.extensions:
                e.extra_link_args += ['-lm']

        super(KivyBuildExt, self).build_extensions()

    def update_if_changed(self, fn, content):
        need_update = True
        if exists(fn):
            with open(fn) as fd:
                need_update = fd.read() != content
        if need_update:
            directory_name = dirname(fn)
            if not exists(directory_name):
                makedirs(directory_name)
            with open(fn, 'w') as fd:
                fd.write(content)
        return need_update


def _check_and_fix_sdl2_mixer(f_path):
    # Between SDL_mixer 2.0.1 and 2.0.4, the included frameworks changed
    # smpeg2 have been replaced with mpg123, but there is no need to fix.
    smpeg2_path = ("{}/Versions/A/Frameworks/smpeg2.framework"
                   "/Versions/A/smpeg2").format(f_path)
    if not exists(smpeg2_path):
        return

    print("Check if SDL2_mixer smpeg2 have an @executable_path")
    rpath_from = ("@executable_path/../Frameworks/SDL2.framework"
                  "/Versions/A/SDL2")
    rpath_to = "@rpath/../../../../SDL2.framework/Versions/A/SDL2"
    output = getoutput(("otool -L '{}'").format(smpeg2_path)).decode('utf-8')
    if "@executable_path" not in output:
        return

    print("WARNING: Your SDL2_mixer version is invalid")
    print("WARNING: The smpeg2 framework embedded in SDL2_mixer contains a")
    print("WARNING: reference to @executable_path that will fail the")
    print("WARNING: execution of your application.")
    print("WARNING: We are going to change:")
    print("WARNING: from: {}".format(rpath_from))
    print("WARNING: to: {}".format(rpath_to))
    getoutput("install_name_tool -change {} {} {}".format(
        rpath_from, rpath_to, smpeg2_path))

    output = getoutput(("otool -L '{}'").format(smpeg2_path))
    if b"@executable_path" not in output:
        print("WARNING: Change successfully applied!")
        print("WARNING: You'll never see this message again.")
    else:
        print("WARNING: Unable to apply the changes, sorry.")


# -----------------------------------------------------------------------------
print("Python path is:\n{}\n".format('\n'.join(sys.path)))
# extract version (simulate doc generation, kivy will be not imported)
environ['KIVY_DOC_INCLUDE'] = '1'
import kivy

# Cython check
# on python-for-android and kivy-ios, cython usage is external
from kivy.tools.packaging.cython_cfg import get_cython_versions, get_cython_msg
CYTHON_REQUIRES_STRING, MIN_CYTHON_STRING, MAX_CYTHON_STRING, \
    CYTHON_UNSUPPORTED = get_cython_versions()
cython_min_msg, cython_max_msg, cython_unsupported_msg = get_cython_msg()

if can_use_cython:
    import Cython
    print('\nFound Cython at', Cython.__file__)

    cy_version_str = Cython.__version__
    cy_ver = LooseVersion(cy_version_str)
    print('Detected supported Cython version {}'.format(cy_version_str))

    if cy_ver < LooseVersion(MIN_CYTHON_STRING):
        print(cython_min_msg)
    elif cy_ver in CYTHON_UNSUPPORTED:
        print(cython_unsupported_msg)
    elif cy_ver > LooseVersion(MAX_CYTHON_STRING):
        print(cython_max_msg)
    sleep(1)

# extra build commands go in the cmdclass dict {'command-name': CommandClass}
# see tools.packaging.{platform}.build.py for custom build commands for
# portable packages. Also e.g. we use build_ext command from cython if its
# installed for c extensions.
from kivy.tools.packaging.factory import FactoryBuild
cmdclass = {
    'build_factory': FactoryBuild,
    'build_ext': KivyBuildExt}

try:
    # add build rules for portable packages to cmdclass
    if platform == 'win32':
        from kivy.tools.packaging.win32.build import WindowsPortableBuild
        cmdclass['build_portable'] = WindowsPortableBuild
    elif platform == 'darwin':
        from kivy.tools.packaging.osx.build import OSXPortableBuild
        cmdclass['build_portable'] = OSXPortableBuild
except ImportError:
    print('User distribution detected, avoid portable command.')

# Detect which opengl version headers to use
if platform in ('android', 'darwin', 'ios', 'rpi', 'mali', 'vc'):
    c_options['use_opengl_es2'] = True
elif c_options['use_opengl_es2'] is None:
    c_options['use_opengl_es2'] = \
        environ.get('KIVY_GRAPHICS', '').lower() == 'gles'

print('Using this graphics system: {}'.format(
    ['OpenGL', 'OpenGL ES 2'][int(c_options['use_opengl_es2'] or False)]))

# check if we are in a kivy-ios build
if platform == 'ios':
    print('Kivy-IOS project environment detect, use it.')
    print('Kivy-IOS project located at {0}'.format(kivy_ios_root))
    c_options['use_ios'] = True
    c_options['use_sdl2'] = True

elif platform == 'android':
    c_options['use_android'] = True

# detect gstreamer, only on desktop
# works if we forced the options or in autodetection
if platform not in ('ios', 'android') and (c_options['use_gstreamer']
                                           in (None, True)):
    gstreamer_valid = False
    if c_options['use_osx_frameworks'] and platform == 'darwin':
        # check the existence of frameworks
        f_path = '/Library/Frameworks/GStreamer.framework'
        if not exists(f_path):
            c_options['use_gstreamer'] = False
            print('GStreamer framework not found, fallback on pkg-config')
        else:
            print('GStreamer framework found')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True
            gst_flags = {
                'extra_link_args': [
                    '-F/Library/Frameworks',
                    '-Xlinker', '-rpath',
                    '-Xlinker', '/Library/Frameworks',
                    '-Xlinker', '-headerpad',
                    '-Xlinker', '190',
                    '-framework', 'GStreamer'],
                'include_dirs': [join(f_path, 'Headers')]}
    elif platform == 'win32':
        gst_flags = pkgconfig('gstreamer-1.0')
        if 'libraries' in gst_flags:
            print('GStreamer found via pkg-config')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True
        else:
            _includes = get_isolated_env_paths()[0] + [get_paths()['include']]
            for include_dir in _includes:
                if exists(join(include_dir, 'gst', 'gst.h')):
                    print('GStreamer found via gst.h')
                    gstreamer_valid = True
                    c_options['use_gstreamer'] = True
                    gst_flags = {
                        'libraries':
                            ['gstreamer-1.0', 'glib-2.0', 'gobject-2.0']}
                    break

    if not gstreamer_valid:
        # use pkg-config approach instead
        gst_flags = pkgconfig('gstreamer-1.0')
        if 'libraries' in gst_flags:
            print('GStreamer found via pkg-config')
            c_options['use_gstreamer'] = True


# detect SDL2, only on desktop and iOS, or android if explicitly enabled
# works if we forced the options or in autodetection
sdl2_flags = {}
if platform == 'win32' and c_options['use_sdl2'] is None:
    c_options['use_sdl2'] = True

if c_options['use_sdl2'] or (
        platform not in ('android',) and c_options['use_sdl2'] is None):

    sdl2_valid = False
    if c_options['use_osx_frameworks'] and platform == 'darwin':
        # check the existence of frameworks
        sdl2_frameworks_search_path = environ.get(
            "KIVY_SDL2_FRAMEWORKS_SEARCH_PATH", "/Library/Frameworks"
        )
        sdl2_valid = True
        sdl2_flags = {
            'extra_link_args': [
                '-F{}'.format(sdl2_frameworks_search_path),
                '-Xlinker', '-rpath',
                '-Xlinker', sdl2_frameworks_search_path,
                '-Xlinker', '-headerpad',
                '-Xlinker', '190'],
            'include_dirs': [],
            'extra_compile_args': ['-F{}'.format(sdl2_frameworks_search_path)]
        }
        for name in ('SDL2', 'SDL2_ttf', 'SDL2_image', 'SDL2_mixer'):
            f_path = '{}/{}.framework'.format(sdl2_frameworks_search_path, name)
            if not exists(f_path):
                print('Missing framework {}'.format(f_path))
                sdl2_valid = False
                continue
            sdl2_flags['extra_link_args'] += ['-framework', name]
            sdl2_flags['include_dirs'] += [join(f_path, 'Headers')]
            print('Found sdl2 frameworks: {}'.format(f_path))
            if name == 'SDL2_mixer':
                _check_and_fix_sdl2_mixer(f_path)

        if not sdl2_valid:
            c_options['use_sdl2'] = False
            print('SDL2 frameworks not found, fallback on pkg-config')
        else:
            c_options['use_sdl2'] = True
            print('Activate SDL2 compilation')

    if not sdl2_valid and platform != "ios":
        # use pkg-config approach instead
        sdl2_flags = pkgconfig('sdl2', 'SDL2_ttf', 'SDL2_image', 'SDL2_mixer')
        if 'libraries' in sdl2_flags:
            print('SDL2 found via pkg-config')
            c_options['use_sdl2'] = True


# -----------------------------------------------------------------------------
# declare flags

def expand(root, *args):
    return join(root, 'kivy', *args)


class CythonExtension(Extension):

    def __init__(self, *args, **kwargs):
        Extension.__init__(self, *args, **kwargs)
        self.cython_directives = {
            'c_string_encoding': 'utf-8',
            'profile': 'USE_PROFILE' in environ,
            'embedsignature': use_embed_signature,
            'language_level': 3,
            'unraisable_tracebacks': True,
        }
        # XXX with pip, setuptools is imported before distutils, and change
        # our pyx to c, then, cythonize doesn't happen. So force again our
        # sources
        self.sources = args[1]


def merge(d1, *args):
    d1 = deepcopy(d1)
    for d2 in args:
        for key, value in d2.items():
            value = deepcopy(value)
            if key in d1:
                d1[key].extend(value)
            else:
                d1[key] = value
    return d1


def determine_base_flags():
    includes, libs = get_isolated_env_paths()

    flags = {
        'libraries': [],
        'include_dirs': [join(src_path, 'kivy', 'include')] + includes,
        'library_dirs': [] + libs,
        'extra_link_args': [],
        'extra_compile_args': []}
    if c_options['use_ios']:
        sysroot = environ.get('IOSSDKROOT', environ.get('SDKROOT'))
        if not sysroot:
            raise Exception('IOSSDKROOT is not set')
        flags['include_dirs'] += [sysroot]
        flags['extra_compile_args'] += ['-isysroot', sysroot]
        flags['extra_link_args'] += ['-isysroot', sysroot]
    elif platform.startswith('freebsd'):
        flags['include_dirs'] += [join(
            environ.get('LOCALBASE', '/usr/local'), 'include')]
        flags['library_dirs'] += [join(
            environ.get('LOCALBASE', '/usr/local'), 'lib')]
    elif platform == 'darwin' and c_options['use_osx_frameworks']:
        v = os.uname()
        if v[2] >= '13.0.0':
            if 'SDKROOT' in environ:
                sysroot = join(environ['SDKROOT'], 'System/Library/Frameworks')
            else:
                # use xcode-select to search on the right Xcode path
                # XXX use the best SDK available instead of a specific one
                import platform as _platform
                xcode_dev = getoutput('xcode-select -p').splitlines()[0]
                sdk_mac_ver = '.'.join(_platform.mac_ver()[0].split('.')[:2])
                print('Xcode detected at {}, and using OS X{} sdk'.format(
                    xcode_dev, sdk_mac_ver))
                sysroot = join(
                    xcode_dev.decode('utf-8'),
                    'Platforms/MacOSX.platform/Developer/SDKs',
                    'MacOSX{}.sdk'.format(sdk_mac_ver),
                    'System/Library/Frameworks')
        else:
            sysroot = ('/System/Library/Frameworks/'
                       'ApplicationServices.framework/Frameworks')
        flags['extra_compile_args'] += ['-F%s' % sysroot]
        flags['extra_link_args'] += ['-F%s' % sysroot]
    elif platform == 'win32':
        flags['include_dirs'] += [get_python_inc(prefix=sys.prefix)]
        flags['library_dirs'] += [join(sys.prefix, "libs")]
    return flags


def determine_gl_flags():
    kivy_graphics_include = join(src_path, 'kivy', 'include')
    flags = {'include_dirs': [kivy_graphics_include], 'libraries': []}
    base_flags = {'include_dirs': [kivy_graphics_include], 'libraries': []}
    cross_sysroot = environ.get('KIVY_CROSS_SYSROOT')

    if c_options['use_opengl_mock']:
        return flags, base_flags
    if platform == 'win32':
        flags['libraries'] = ['opengl32', 'glew32']
    elif platform == 'ios':
        flags['libraries'] = ['GLESv2']
        flags['extra_link_args'] = ['-framework', 'OpenGLES']
    elif platform == 'darwin':
        flags['extra_link_args'] = ['-framework', 'OpenGL']
    elif platform.startswith('freebsd'):
        flags['libraries'] = ['GL']
    elif platform.startswith('openbsd'):
        flags['include_dirs'] = ['/usr/X11R6/include']
        flags['library_dirs'] = ['/usr/X11R6/lib']
        flags['libraries'] = ['GL']
    elif platform == 'android':
        flags['include_dirs'] = [join(ndkplatform, 'usr', 'include')]
        flags['library_dirs'] = [join(ndkplatform, 'usr', 'lib')]
        flags['libraries'] = ['GLESv2']
    elif platform == 'rpi':

        if not cross_sysroot:
            flags['include_dirs'] = [
                '/opt/vc/include',
                '/opt/vc/include/interface/vcos/pthreads',
                '/opt/vc/include/interface/vmcs_host/linux']
            flags['library_dirs'] = ['/opt/vc/lib']
            brcm_lib_files = (
                '/opt/vc/lib/libbrcmEGL.so',
                '/opt/vc/lib/libbrcmGLESv2.so')

        else:
            print("KIVY_CROSS_SYSROOT: " + cross_sysroot)
            flags['include_dirs'] = [
                cross_sysroot + '/usr/include',
                cross_sysroot + '/usr/include/interface/vcos/pthreads',
                cross_sysroot + '/usr/include/interface/vmcs_host/linux']
            flags['library_dirs'] = [cross_sysroot + '/usr/lib']
            brcm_lib_files = (
                cross_sysroot + '/usr/lib/libbrcmEGL.so',
                cross_sysroot + '/usr/lib/libbrcmGLESv2.so')

        if all((exists(lib) for lib in brcm_lib_files)):
            print('Found brcmEGL and brcmGLES library files '
                  'for rpi platform at ' + dirname(brcm_lib_files[0]))
            gl_libs = ['brcmEGL', 'brcmGLESv2']
        else:
            print(
                'Failed to find brcmEGL and brcmGLESv2 library files '
                'for rpi platform, falling back to EGL and GLESv2.')
            gl_libs = ['EGL', 'GLESv2']
        flags['libraries'] = ['bcm_host'] + gl_libs
    elif platform in ['mali', 'vc']:
        flags['include_dirs'] = ['/usr/include/']
        flags['library_dirs'] = ['/usr/lib/arm-linux-gnueabihf']
        flags['libraries'] = ['GLESv2']
        c_options['use_x11'] = True
        c_options['use_egl'] = True
    else:
        flags['libraries'] = ['GL']
    return flags, base_flags


def determine_sdl2():
    flags = {}
    if not c_options['use_sdl2']:
        return flags

    sdl2_path = environ.get('KIVY_SDL2_PATH', None)

    if sdl2_flags and not sdl2_path and platform == 'darwin':
        return sdl2_flags

    includes, _ = get_isolated_env_paths()

    # no pkgconfig info, or we want to use a specific sdl2 path, so perform
    # manual configuration
    flags['libraries'] = ['SDL2', 'SDL2_ttf', 'SDL2_image', 'SDL2_mixer']
    split_chr = ';' if platform == 'win32' else ':'
    sdl2_paths = sdl2_path.split(split_chr) if sdl2_path else []

    if not sdl2_paths:
        sdl2_paths = []
        for include in includes + [join(sys.prefix, 'include')]:
            sdl_inc = join(include, 'SDL2')
            if isdir(sdl_inc):
                sdl2_paths.append(sdl_inc)
        sdl2_paths.extend(['/usr/local/include/SDL2', '/usr/include/SDL2'])

    flags['include_dirs'] = sdl2_paths
    flags['extra_link_args'] = []
    flags['extra_compile_args'] = []
    flags['library_dirs'] = (
        sdl2_paths if sdl2_paths else
        ['/usr/local/lib/'])

    if sdl2_flags:
        flags = merge(flags, sdl2_flags)

    # ensure headers for all the SDL2 and sub libraries are available
    libs_to_check = ['SDL', 'SDL_mixer', 'SDL_ttf', 'SDL_image']
    can_compile = True
    for lib in libs_to_check:
        found = False
        for d in flags['include_dirs']:
            fn = join(d, '{}.h'.format(lib))
            if exists(fn):
                found = True
                print('SDL2: found {} header at {}'.format(lib, fn))
                break

        if not found:
            print('SDL2: missing sub library {}'.format(lib))
            can_compile = False

    if not can_compile:
        c_options['use_sdl2'] = False
        return {}

    return flags


base_flags = determine_base_flags()
gl_flags, gl_flags_base = determine_gl_flags()

# -----------------------------------------------------------------------------
# sources to compile
# all the dependencies have been found manually with:
# grep -inr -E '(cimport|include)' kivy/graphics/context_instructions.{pxd,pyx}
graphics_dependencies = {
    'buffer.pyx': ['common.pxi'],
    'context.pxd': ['instructions.pxd', 'texture.pxd', 'vbo.pxd', 'cgl.pxd'],
    'cgl.pxd': ['common.pxi', 'config.pxi', 'gl_redirect.h'],
    'compiler.pxd': ['instructions.pxd'],
    'compiler.pyx': ['context_instructions.pxd'],
    'cgl.pyx': ['cgl.pxd'],
    'cgl_mock.pyx': ['cgl.pxd'],
    'cgl_sdl2.pyx': ['cgl.pxd'],
    'cgl_gl.pyx': ['cgl.pxd'],
    'cgl_glew.pyx': ['cgl.pxd'],
    'context_instructions.pxd': [
        'transformation.pxd', 'instructions.pxd', 'texture.pxd'],
    'fbo.pxd': ['cgl.pxd', 'instructions.pxd', 'texture.pxd'],
    'fbo.pyx': [
        'config.pxi', 'opcodes.pxi', 'transformation.pxd', 'context.pxd'],
    'gl_instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd', 'instructions.pxd'],
    'instructions.pxd': [
        'vbo.pxd', 'context_instructions.pxd', 'compiler.pxd', 'shader.pxd',
        'texture.pxd', '../_event.pxd'],
    'instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd',
        'context.pxd', 'common.pxi', 'vertex.pxd', 'transformation.pxd'],
    'opengl.pyx': [
        'config.pxi', 'common.pxi', 'cgl.pxd', 'gl_redirect.h'],
    'opengl_utils.pyx': [
        'opengl_utils_def.pxi', 'cgl.pxd', ],
    'shader.pxd': ['cgl.pxd', 'transformation.pxd', 'vertex.pxd'],
    'shader.pyx': [
        'config.pxi', 'common.pxi', 'cgl.pxd',
        'vertex.pxd', 'transformation.pxd', 'context.pxd',
        'gl_debug_logger.pxi'],
    'stencil_instructions.pxd': ['instructions.pxd'],
    'stencil_instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd',
        'gl_debug_logger.pxi'],
    'scissor_instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd'],
    'svg.pyx': ['config.pxi', 'common.pxi', 'texture.pxd', 'instructions.pxd',
                'vertex_instructions.pxd', 'tesselator.pxd'],
    'texture.pxd': ['cgl.pxd'],
    'texture.pyx': [
        'config.pxi', 'common.pxi', 'opengl_utils_def.pxi', 'context.pxd',
        'cgl.pxd', 'opengl_utils.pxd',
        'img_tools.pxi', 'gl_debug_logger.pxi'],
    'vbo.pxd': ['buffer.pxd', 'cgl.pxd', 'vertex.pxd'],
    'vbo.pyx': [
        'config.pxi', 'common.pxi', 'context.pxd',
        'instructions.pxd', 'shader.pxd', 'gl_debug_logger.pxi'],
    'vertex.pxd': ['cgl.pxd'],
    'vertex.pyx': ['config.pxi', 'common.pxi'],
    'vertex_instructions.pyx': [
        'config.pxi', 'common.pxi', 'vbo.pxd', 'vertex.pxd',
        'instructions.pxd', 'vertex_instructions.pxd',
        'cgl.pxd', 'texture.pxd', 'vertex_instructions_line.pxi'],
    'vertex_instructions_line.pxi': ['stencil_instructions.pxd']}

sources = {
    '_event.pyx': merge(base_flags, {'depends': ['properties.pxd']}),
    '_clock.pyx': {},
    'weakproxy.pyx': {},
    'properties.pyx': merge(
        base_flags, {'depends': ['_event.pxd', '_metrics.pxd']}),
    '_metrics.pyx': merge(base_flags, {'depends': ['_event.pxd']}),
    'graphics/buffer.pyx': merge(base_flags, gl_flags_base),
    'graphics/context.pyx': merge(base_flags, gl_flags_base),
    'graphics/compiler.pyx': merge(base_flags, gl_flags_base),
    'graphics/context_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/fbo.pyx': merge(base_flags, gl_flags_base),
    'graphics/gl_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/opengl.pyx': merge(base_flags, gl_flags_base),
    'graphics/opengl_utils.pyx': merge(base_flags, gl_flags_base),
    'graphics/shader.pyx': merge(base_flags, gl_flags_base),
    'graphics/stencil_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/scissor_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/texture.pyx': merge(base_flags, gl_flags_base),
    'graphics/transformation.pyx': merge(base_flags, gl_flags_base),
    'graphics/vbo.pyx': merge(base_flags, gl_flags_base),
    'graphics/vertex.pyx': merge(base_flags, gl_flags_base),
    'graphics/vertex_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl_backend/cgl_mock.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl_backend/cgl_gl.pyx': merge(base_flags, gl_flags),
    'graphics/cgl_backend/cgl_glew.pyx': merge(base_flags, gl_flags),
    'graphics/cgl_backend/cgl_sdl2.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl_backend/cgl_debug.pyx': merge(base_flags, gl_flags_base),
    'core/text/text_layout.pyx': base_flags,
    'core/window/window_info.pyx': base_flags,
    'graphics/tesselator.pyx': merge(base_flags, {
        'include_dirs': ['kivy/lib/libtess2/Include'],
        'c_depends': [
            'lib/libtess2/Source/bucketalloc.c',
            'lib/libtess2/Source/dict.c',
            'lib/libtess2/Source/geom.c',
            'lib/libtess2/Source/mesh.c',
            'lib/libtess2/Source/priorityq.c',
            'lib/libtess2/Source/sweep.c',
            'lib/libtess2/Source/tess.c'
        ]
    }),
    'graphics/svg.pyx': merge(base_flags, gl_flags_base)
}

if c_options["use_sdl2"]:
    sdl2_flags = determine_sdl2()

if c_options['use_sdl2'] and sdl2_flags:
    sources['graphics/cgl_backend/cgl_sdl2.pyx'] = merge(
        sources['graphics/cgl_backend/cgl_sdl2.pyx'], sdl2_flags)
    sdl2_depends = {'depends': ['lib/sdl2.pxi']}
    for source_file in ('core/window/_window_sdl2.pyx',
                        'core/image/_img_sdl2.pyx',
                        'core/text/_text_sdl2.pyx',
                        'core/audio/audio_sdl2.pyx',
                        'core/clipboard/_clipboard_sdl2.pyx'):
        sources[source_file] = merge(
            base_flags, sdl2_flags, sdl2_depends)

if c_options['use_pangoft2'] in (None, True) and platform not in (
                                      'android', 'ios', 'win32'):
    pango_flags = pkgconfig('pangoft2')
    if pango_flags and 'libraries' in pango_flags:
        print('Pango: pangoft2 found via pkg-config')
        c_options['use_pangoft2'] = True
        pango_depends = {'depends': [
            'lib/pango/pangoft2.pxi',
            'lib/pango/pangoft2.h']}
        sources['core/text/_text_pango.pyx'] = merge(
                base_flags, pango_flags, pango_depends)
        print(sources['core/text/_text_pango.pyx'])

if platform in ('darwin', 'ios'):
    # activate ImageIO provider for our core image
    if platform == 'ios':
        osx_flags = {'extra_link_args': [
            '-framework', 'Foundation',
            '-framework', 'UIKit',
            '-framework', 'AudioToolbox',
            '-framework', 'CoreGraphics',
            '-framework', 'QuartzCore',
            '-framework', 'ImageIO',
            '-framework', 'Accelerate']}
    else:
        osx_flags = {'extra_link_args': [
            '-framework', 'ApplicationServices']}
    sources['core/image/img_imageio.pyx'] = merge(
        base_flags, osx_flags)

if c_options['use_avfoundation']:
    import platform as _platform
    mac_ver = [int(x) for x in _platform.mac_ver()[0].split('.')[:2]]
    if mac_ver >= [10, 7] or platform == 'ios':
        osx_flags = {
            'extra_link_args': ['-framework', 'AVFoundation'],
            'extra_compile_args': ['-ObjC++'],
            'depends': ['core/camera/camera_avfoundation_implem.m']}
        sources['core/camera/camera_avfoundation.pyx'] = merge(
            base_flags, osx_flags)
    else:
        print('AVFoundation cannot be used, OSX >= 10.7 is required')

if c_options['use_rpi']:
    sources['lib/vidcore_lite/egl.pyx'] = merge(
        base_flags, gl_flags)
    sources['lib/vidcore_lite/bcm.pyx'] = merge(
        base_flags, gl_flags)

if c_options['use_x11']:
    libs = ['Xrender', 'X11']
    if c_options['use_egl']:
        libs += ['EGL']
    else:
        libs += ['GL']
    sources['core/window/window_x11.pyx'] = merge(
        base_flags, gl_flags, {
            # FIXME add an option to depend on them but not compile them
            # cause keytab is included in core, and core is included in
            # window_x11
            #
            # 'depends': [
            #    'core/window/window_x11_keytab.c',
            #    'core/window/window_x11_core.c'],
            'libraries': libs})

if c_options['use_gstreamer']:
    sources['lib/gstplayer/_gstplayer.pyx'] = merge(
        base_flags, gst_flags, {
            'depends': ['lib/gstplayer/_gstplayer.h']})

# -----------------------------------------------------------------------------
# extension modules


def get_dependencies(name, deps=None):
    if deps is None:
        deps = []
    for dep in graphics_dependencies.get(name, []):
        if dep not in deps:
            deps.append(dep)
            get_dependencies(dep, deps)
    return deps


def resolve_dependencies(fn, depends):
    fn = basename(fn)
    deps = []
    get_dependencies(fn, deps)
    get_dependencies(fn.replace('.pyx', '.pxd'), deps)

    deps_final = []
    paths_to_test = ['graphics', 'include']
    for dep in deps:
        found = False
        for path in paths_to_test:
            filename = expand(src_path, path, dep)
            if exists(filename):
                deps_final.append(filename)
                found = True
                break
        if not found:
            print('ERROR: Dependency for {} not resolved: {}'.format(
                fn, dep
            ))

    return deps_final


def get_extensions_from_sources(sources):
    ext_modules = []
    if environ.get('KIVY_FAKE_BUILDEXT'):
        print('Fake build_ext asked, will generate only .h/.c')
        return ext_modules
    for pyx, flags in sources.items():
        is_graphics = pyx.startswith('graphics')
        pyx_path = expand(src_path, pyx)
        depends = [expand(src_path, x) for x in flags.pop('depends', [])]
        c_depends = [expand(src_path, x) for x in flags.pop('c_depends', [])]
        if not can_use_cython:
            # can't use cython, so use the .c files instead.
            pyx_path = '%s.c' % pyx_path[:-4]
        if is_graphics:
            depends = resolve_dependencies(pyx_path, depends)
        f_depends = [x for x in depends if x.rsplit('.', 1)[-1] in (
            'c', 'cpp', 'm')]
        module_name = '.'.join(['kivy'] + pyx[:-4].split('/'))
        flags_clean = {'depends': depends}
        for key, value in flags.items():
            if len(value):
                flags_clean[key] = value
        ext_modules.append(CythonExtension(
            module_name, [pyx_path] + f_depends + c_depends, **flags_clean))
    return ext_modules


ext_modules = get_extensions_from_sources(sources)


# -----------------------------------------------------------------------------
# automatically detect data files
split_examples = int(environ.get('KIVY_SPLIT_EXAMPLES', '0'))
data_file_prefix = 'share/kivy-'
examples = {}
examples_allowed_ext = ('readme', 'py', 'wav', 'png', 'jpg', 'svg', 'json',
                        'avi', 'gif', 'txt', 'ttf', 'obj', 'mtl', 'kv', 'mpg',
                        'glsl', 'zip')
for root, subFolders, files in walk('examples'):
    for fn in files:
        ext = fn.split('.')[-1].lower()
        if ext not in examples_allowed_ext:
            continue
        filename = join(root, fn)
        directory = '%s%s' % (data_file_prefix, dirname(filename))
        if directory not in examples:
            examples[directory] = []
        examples[directory].append(filename)

binary_deps = []
binary_deps_path = join(src_path, 'kivy', 'binary_deps')
if isdir(binary_deps_path):
    for root, dirnames, filenames in walk(binary_deps_path):
        for fname in filenames:
            binary_deps.append(
                join(root.replace(binary_deps_path, 'binary_deps'), fname))


def glob_paths(*patterns, excludes=('.pyc', )):
    files = []
    base = Path(join(src_path, 'kivy'))

    for pat in patterns:
        for f in base.glob(pat):
            if f.suffix in excludes:
                continue
            files.append(str(f.relative_to(base)))
    return files


# -----------------------------------------------------------------------------
# setup !
if not build_examples:
    setup(
        name='Kivy',
        version=__version__,
        author='Kivy Team and other contributors',
        author_email='kivy-dev@googlegroups.com',
        url='http://kivy.org',
        project_urls={
            'Source': 'https://github.com/kivy/kivy',
        },
        license='MIT',
        description=(
            'A software library for rapid development of '
            'hardware-accelerated multitouch applications.'),
        long_description=get_description(),
        long_description_content_type='text/markdown',
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        packages=find_packages(include=['kivy*']),
        package_dir={'kivy': 'kivy'},
        package_data={
            'kivy':
                glob_paths('*.pxd', '*.pxi') +
                glob_paths('**/*.pxd', '**/*.pxi') +
                glob_paths('data/**/*.*') +
                glob_paths('include/**/*.*') +
                glob_paths('tools/**/*.*', excludes=('.pyc', '.enc')) +
                glob_paths('graphics/**/*.h') +
                glob_paths('tests/**/*.*') +
                [
                    'setupconfig.py',
                ] + binary_deps
        },
        data_files=[] if split_examples else list(examples.items()),
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: MacOS X',
            'Environment :: Win32 (MS Windows)',
            'Environment :: X11 Applications',
            'Intended Audience :: Developers',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: BSD :: FreeBSD',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Topic :: Artistic Software',
            'Topic :: Games/Entertainment',
            'Topic :: Multimedia :: Graphics :: 3D Rendering',
            'Topic :: Multimedia :: Graphics :: Capture :: Digital Camera',
            'Topic :: Multimedia :: Graphics :: Presentation',
            'Topic :: Multimedia :: Graphics :: Viewers',
            'Topic :: Multimedia :: Sound/Audio :: Players :: MP3',
            'Topic :: Multimedia :: Video :: Display',
            'Topic :: Scientific/Engineering :: Human Machine Interfaces',
            'Topic :: Scientific/Engineering :: Visualization',
            ('Topic :: Software Development :: Libraries :: '
             'Application Frameworks'),
            'Topic :: Software Development :: User Interfaces'])
else:
    setup(
        name='Kivy-examples',
        version=__version__,
        author='Kivy Team and other contributors',
        author_email='kivy-dev@googlegroups.com',
        url='http://kivy.org',
        license='MIT',
        description=('Kivy examples.'),
        long_description_content_type='text/markdown',
        long_description=get_description(),
        data_files=list(examples.items()))



"""
CGL: standard C interface for OpenGL
====================================

Kivy uses OpenGL and therefore requires a backend that provides it.
The backend used is controlled through the ``USE_OPENGL_MOCK`` and ``USE_SDL2``
compile-time variables and through the ``KIVY_GL_BACKEND`` runtime
environmental variable.

Currently, OpenGL is used through direct linking (gl/glew), sdl2,
or by mocking it. Setting ``USE_OPENGL_MOCK`` disables gl/glew.
Similarly, setting ``USE_SDL2`` to ``0`` will disable sdl2. Mocking
is always available.

At runtime the following backends are available and can be set using
``KIVY_GL_BACKEND``:

* ``gl`` -- Available on unix (the default backend). Unavailable when
  ``USE_OPENGL_MOCK=0``. Requires gl be installed.
* ``glew`` -- Available on Windows (the default backend). Unavailable when
  ``USE_OPENGL_MOCK=0``. Requires glew be installed.
* ``sdl2`` -- Available on Windows/unix (the default when gl/glew is disabled).
  Unavailable when ``USE_SDL2=0``. Requires ``kivy_deps.sdl2`` be installed.
* ``angle_sdl2`` -- Available on Windows with Python 3.5+.
  Unavailable when ``USE_SDL2=0``. Requires ``kivy_deps.sdl2`` and
  ``kivy_deps.angle`` be installed.
* ``mock`` -- Always available. Doesn't actually do anything.


Additionally, the following environmental runtime variables control the graphics
system:

* ``KIVY_GL_DEBUG`` -- Logs al gl calls when ``1``.
* ``KIVY_GRAPHICS`` -- Forces OpenGL ES2 when it is ``gles``. OpenGL ES2 is always
  used on the android, ios, rpi, and mali OSs.
"""

include "../include/config.pxi"

from sys import platform
from os import environ
from .cgl cimport GLES2_Context
import importlib
from kivy.logger import Logger
from threading import get_ident
from kivy.config import Config

cdef GLES2_Context g_cgl
cdef GLES2_Context *cgl = &g_cgl
cdef object cgl_name = None
cdef int kivy_opengl_es2 = USE_OPENGL_ES2 or environ.get('KIVY_GRAPHICS', '').lower() == 'gles'
cdef unsigned long initialized_tid = 0
cdef public int verify_gl_main_thread = 1
"""Whether we should check if the gl instructions occur in a thread outside the main thread.
"""


def _update_verify_gl_main_thread(*args):
    global verify_gl_main_thread
    verify_gl_main_thread = Config.getboolean('graphics', 'verify_gl_main_thread')

if not environ.get('KIVY_DOC_INCLUDE'):
    Config.add_callback(_update_verify_gl_main_thread, 'graphics', 'verify_gl_main_thread')
    _update_verify_gl_main_thread()


cpdef cgl_get_initialized_backend_name():
    return cgl_name


cpdef cgl_get_backend_name(allowed=[], ignored=[]):
    if cgl_name:
        return cgl_name
    name = environ.get("KIVY_GL_BACKEND")
    if name:
        return name.lower()

    for name in ('glew', 'sdl2', 'gl', 'mock'):
        if allowed and name not in allowed:
            continue
        if name in ignored:
            continue

        mod = importlib.import_module("kivy.graphics.cgl_backend.cgl_{}".format(name))
        if mod.is_backend_supported():
            return name
    assert False


cdef GLES2_Context *cgl_get_context():
    return cgl


cdef void cgl_set_context(GLES2_Context* ctx):
    global cgl
    cgl = ctx


cpdef cgl_init(allowed=[], ignored=[]):
    Logger.info('GL: Using the "{}" graphics system'.format(
        'OpenGL ES 2' if kivy_opengl_es2 else 'OpenGL'))
    global cgl_name, initialized_tid
    cgl_name = backend = cgl_get_backend_name(allowed, ignored)
    initialized_tid = get_ident()

    # for ANGLE, currently we use sdl2, and only on windows.
    if backend == "angle_sdl2":
        if platform != "win32":
            raise Exception("CGL: ANGLE backend can be used only on Windows")
        backend = "sdl2"

    if cgl_name not in {'glew', 'sdl2', 'angle_sdl2', 'mock', 'gl'}:
        raise ValueError('{} is not a recognized GL backend'.format(backend))

    mod = importlib.import_module("kivy.graphics.cgl_backend.cgl_{}".format(backend))
    mod.init_backend()
    log_cgl_funcs()

    use_debug = environ.get("KIVY_GL_DEBUG") == "1"
    if use_debug:
        mod = importlib.import_module("kivy.graphics.cgl_backend.cgl_debug")
        mod.init_backend_debug()


cdef void log_cgl_funcs() except *:
    if cgl.glActiveTexture == NULL:
        Logger.debug('GL: glActiveTexture is not available')
    if cgl.glAttachShader == NULL:
        Logger.debug('GL: glAttachShader is not available')
    if cgl.glBindAttribLocation == NULL:
        Logger.debug('GL: glBindAttribLocation is not available')
    if cgl.glBindBuffer == NULL:
        Logger.debug('GL: glBindBuffer is not available')
    if cgl.glBindFramebuffer == NULL:
        Logger.debug('GL: glBindFramebuffer is not available')
    if cgl.glBindRenderbuffer == NULL:
        Logger.debug('GL: glBindRenderbuffer is not available')
    if cgl.glBindTexture == NULL:
        Logger.debug('GL: glBindTexture is not available')
    if cgl.glBlendColor == NULL:
        Logger.debug('GL: glBlendColor is not available')
    if cgl.glBlendEquation == NULL:
        Logger.debug('GL: glBlendEquation is not available')
    if cgl.glBlendEquationSeparate == NULL:
        Logger.debug('GL: glBlendEquationSeparate is not available')
    if cgl.glBlendFunc == NULL:
        Logger.debug('GL: glBlendFunc is not available')
    if cgl.glBlendFuncSeparate == NULL:
        Logger.debug('GL: glBlendFuncSeparate is not available')
    if cgl.glBufferData == NULL:
        Logger.debug('GL: glBufferData is not available')
    if cgl.glBufferSubData == NULL:
        Logger.debug('GL: glBufferSubData is not available')
    if cgl.glCheckFramebufferStatus == NULL:
        Logger.debug('GL: glCheckFramebufferStatus is not available')
    if cgl.glClear == NULL:
        Logger.debug('GL: glClear is not available')
    if cgl.glClearColor == NULL:
        Logger.debug('GL: glClearColor is not available')
    if cgl.glClearStencil == NULL:
        Logger.debug('GL: glClearStencil is not available')
    if cgl.glColorMask == NULL:
        Logger.debug('GL: glColorMask is not available')
    if cgl.glCompileShader == NULL:
        Logger.debug('GL: glCompileShader is not available')
    if cgl.glCompressedTexImage2D == NULL:
        Logger.debug('GL: glCompressedTexImage2D is not available')
    if cgl.glCompressedTexSubImage2D == NULL:
        Logger.debug('GL: glCompressedTexSubImage2D is not available')
    if cgl.glCopyTexImage2D == NULL:
        Logger.debug('GL: glCopyTexImage2D is not available')
    if cgl.glCopyTexSubImage2D == NULL:
        Logger.debug('GL: glCopyTexSubImage2D is not available')
    if cgl.glCreateProgram == NULL:
        Logger.debug('GL: glCreateProgram is not available')
    if cgl.glCreateShader == NULL:
        Logger.debug('GL: glCreateShader is not available')
    if cgl.glCullFace == NULL:
        Logger.debug('GL: glCullFace is not available')
    if cgl.glDeleteBuffers == NULL:
        Logger.debug('GL: glDeleteBuffers is not available')
    if cgl.glDeleteFramebuffers == NULL:
        Logger.debug('GL: glDeleteFramebuffers is not available')
    if cgl.glDeleteProgram == NULL:
        Logger.debug('GL: glDeleteProgram is not available')
    if cgl.glDeleteRenderbuffers == NULL:
        Logger.debug('GL: glDeleteRenderbuffers is not available')
    if cgl.glDeleteShader == NULL:
        Logger.debug('GL: glDeleteShader is not available')
    if cgl.glDeleteTextures == NULL:
        Logger.debug('GL: glDeleteTextures is not available')
    if cgl.glDepthFunc == NULL:
        Logger.debug('GL: glDepthFunc is not available')
    if cgl.glDepthMask == NULL:
        Logger.debug('GL: glDepthMask is not available')
    if cgl.glDetachShader == NULL:
        Logger.debug('GL: glDetachShader is not available')
    if cgl.glDisable == NULL:
        Logger.debug('GL: glDisable is not available')
    if cgl.glDisableVertexAttribArray == NULL:
        Logger.debug('GL: glDisableVertexAttribArray is not available')
    if cgl.glDrawArrays == NULL:
        Logger.debug('GL: glDrawArrays is not available')
    if cgl.glDrawElements == NULL:
        Logger.debug('GL: glDrawElements is not available')
    if cgl.glEnable == NULL:
        Logger.debug('GL: glEnable is not available')
    if cgl.glEnableVertexAttribArray == NULL:
        Logger.debug('GL: glEnableVertexAttribArray is not available')
    if cgl.glFinish == NULL:
        Logger.debug('GL: glFinish is not available')
    if cgl.glFlush == NULL:
        Logger.debug('GL: glFlush is not available')
    if cgl.glFramebufferRenderbuffer == NULL:
        Logger.debug('GL: glFramebufferRenderbuffer is not available')
    if cgl.glFramebufferTexture2D == NULL:
        Logger.debug('GL: glFramebufferTexture2D is not available')
    if cgl.glFrontFace == NULL:
        Logger.debug('GL: glFrontFace is not available')
    if cgl.glGenBuffers == NULL:
        Logger.debug('GL: glGenBuffers is not available')
    if cgl.glGenerateMipmap == NULL:
        Logger.debug('GL: glGenerateMipmap is not available')
    if cgl.glGenFramebuffers == NULL:
        Logger.debug('GL: glGenFramebuffers is not available')
    if cgl.glGenRenderbuffers == NULL:
        Logger.debug('GL: glGenRenderbuffers is not available')
    if cgl.glGenTextures == NULL:
        Logger.debug('GL: glGenTextures is not available')
    if cgl.glGetActiveAttrib == NULL:
        Logger.debug('GL: glGetActiveAttrib is not available')
    if cgl.glGetActiveUniform == NULL:
        Logger.debug('GL: glGetActiveUniform is not available')
    if cgl.glGetAttachedShaders == NULL:
        Logger.debug('GL: glGetAttachedShaders is not available')
    if cgl.glGetAttribLocation == NULL:
        Logger.debug('GL: glGetAttribLocation is not available')
    if cgl.glGetBooleanv == NULL:
        Logger.debug('GL: glGetBooleanv is not available')
    if cgl.glGetBufferParameteriv == NULL:
        Logger.debug('GL: glGetBufferParameteriv is not available')
    if cgl.glGetError == NULL:
        Logger.debug('GL: glGetError is not available')
    if cgl.glGetFloatv == NULL:
        Logger.debug('GL: glGetFloatv is not available')
    if cgl.glGetFramebufferAttachmentParameteriv == NULL:
        Logger.debug('GL: glGetFramebufferAttachmentParameteriv is not available')
    if cgl.glGetIntegerv == NULL:
        Logger.debug('GL: glGetIntegerv is not available')
    if cgl.glGetProgramInfoLog == NULL:
        Logger.debug('GL: glGetProgramInfoLog is not available')
    if cgl.glGetProgramiv == NULL:
        Logger.debug('GL: glGetProgramiv is not available')
    if cgl.glGetRenderbufferParameteriv == NULL:
        Logger.debug('GL: glGetRenderbufferParameteriv is not available')
    if cgl.glGetShaderInfoLog == NULL:
        Logger.debug('GL: glGetShaderInfoLog is not available')
    if cgl.glGetShaderiv == NULL:
        Logger.debug('GL: glGetShaderiv is not available')
    if cgl.glGetShaderSource == NULL:
        Logger.debug('GL: glGetShaderSource is not available')
    if cgl.glGetString == NULL:
        Logger.debug('GL: glGetString is not available')
    if cgl.glGetTexParameterfv == NULL:
        Logger.debug('GL: glGetTexParameterfv is not available')
    if cgl.glGetTexParameteriv == NULL:
        Logger.debug('GL: glGetTexParameteriv is not available')
    if cgl.glGetUniformfv == NULL:
        Logger.debug('GL: glGetUniformfv is not available')
    if cgl.glGetUniformiv == NULL:
        Logger.debug('GL: glGetUniformiv is not available')
    if cgl.glGetUniformLocation == NULL:
        Logger.debug('GL: glGetUniformLocation is not available')
    if cgl.glGetVertexAttribfv == NULL:
        Logger.debug('GL: glGetVertexAttribfv is not available')
    if cgl.glGetVertexAttribiv == NULL:
        Logger.debug('GL: glGetVertexAttribiv is not available')
    if cgl.glHint == NULL:
        Logger.debug('GL: glHint is not available')
    if cgl.glIsBuffer == NULL:
        Logger.debug('GL: glIsBuffer is not available')
    if cgl.glIsEnabled == NULL:
        Logger.debug('GL: glIsEnabled is not available')
    if cgl.glIsFramebuffer == NULL:
        Logger.debug('GL: glIsFramebuffer is not available')
    if cgl.glIsProgram == NULL:
        Logger.debug('GL: glIsProgram is not available')
    if cgl.glIsRenderbuffer == NULL:
        Logger.debug('GL: glIsRenderbuffer is not available')
    if cgl.glIsShader == NULL:
        Logger.debug('GL: glIsShader is not available')
    if cgl.glIsTexture == NULL:
        Logger.debug('GL: glIsTexture is not available')
    if cgl.glLineWidth == NULL:
        Logger.debug('GL: glLineWidth is not available')
    if cgl.glLinkProgram == NULL:
        Logger.debug('GL: glLinkProgram is not available')
    if cgl.glPixelStorei == NULL:
        Logger.debug('GL: glPixelStorei is not available')
    if cgl.glPolygonOffset == NULL:
        Logger.debug('GL: glPolygonOffset is not available')
    if cgl.glReadPixels == NULL:
        Logger.debug('GL: glReadPixels is not available')
    if cgl.glRenderbufferStorage == NULL:
        Logger.debug('GL: glRenderbufferStorage is not available')
    if cgl.glSampleCoverage == NULL:
        Logger.debug('GL: glSampleCoverage is not available')
    if cgl.glScissor == NULL:
        Logger.debug('GL: glScissor is not available')
    if cgl.glShaderBinary == NULL:
        Logger.debug('GL: glShaderBinary is not available')
    if cgl.glShaderSource == NULL:
        Logger.debug('GL: glShaderSource is not available')
    if cgl.glStencilFunc == NULL:
        Logger.debug('GL: glStencilFunc is not available')
    if cgl.glStencilFuncSeparate == NULL:
        Logger.debug('GL: glStencilFuncSeparate is not available')
    if cgl.glStencilMask == NULL:
        Logger.debug('GL: glStencilMask is not available')
    if cgl.glStencilMaskSeparate == NULL:
        Logger.debug('GL: glStencilMaskSeparate is not available')
    if cgl.glStencilOp == NULL:
        Logger.debug('GL: glStencilOp is not available')
    if cgl.glStencilOpSeparate == NULL:
        Logger.debug('GL: glStencilOpSeparate is not available')
    if cgl.glTexImage2D == NULL:
        Logger.debug('GL: glTexImage2D is not available')
    if cgl.glTexParameterf == NULL:
        Logger.debug('GL: glTexParameterf is not available')
    if cgl.glTexParameteri == NULL:
        Logger.debug('GL: glTexParameteri is not available')
    if cgl.glTexSubImage2D == NULL:
        Logger.debug('GL: glTexSubImage2D is not available')
    if cgl.glUniform1f == NULL:
        Logger.debug('GL: glUniform1f is not available')
    if cgl.glUniform1fv == NULL:
        Logger.debug('GL: glUniform1fv is not available')
    if cgl.glUniform1i == NULL:
        Logger.debug('GL: glUniform1i is not available')
    if cgl.glUniform1iv == NULL:
        Logger.debug('GL: glUniform1iv is not available')
    if cgl.glUniform2f == NULL:
        Logger.debug('GL: glUniform2f is not available')
    if cgl.glUniform2fv == NULL:
        Logger.debug('GL: glUniform2fv is not available')
    if cgl.glUniform2i == NULL:
        Logger.debug('GL: glUniform2i is not available')
    if cgl.glUniform2iv == NULL:
        Logger.debug('GL: glUniform2iv is not available')
    if cgl.glUniform3f == NULL:
        Logger.debug('GL: glUniform3f is not available')
    if cgl.glUniform3fv == NULL:
        Logger.debug('GL: glUniform3fv is not available')
    if cgl.glUniform3i == NULL:
        Logger.debug('GL: glUniform3i is not available')
    if cgl.glUniform3iv == NULL:
        Logger.debug('GL: glUniform3iv is not available')
    if cgl.glUniform4f == NULL:
        Logger.debug('GL: glUniform4f is not available')
    if cgl.glUniform4fv == NULL:
        Logger.debug('GL: glUniform4fv is not available')
    if cgl.glUniform4i == NULL:
        Logger.debug('GL: glUniform4i is not available')
    if cgl.glUniform4iv == NULL:
        Logger.debug('GL: glUniform4iv is not available')
    if cgl.glUniformMatrix4fv == NULL:
        Logger.debug('GL: glUniformMatrix4fv is not available')
    if cgl.glUseProgram == NULL:
        Logger.debug('GL: glUseProgram is not available')
    if cgl.glValidateProgram == NULL:
        Logger.debug('GL: glValidateProgram is not available')
    if cgl.glVertexAttrib1f == NULL:
        Logger.debug('GL: glVertexAttrib1f is not available')
    if cgl.glVertexAttrib2f == NULL:
        Logger.debug('GL: glVertexAttrib2f is not available')
    if cgl.glVertexAttrib3f == NULL:
        Logger.debug('GL: glVertexAttrib3f is not available')
    if cgl.glVertexAttrib4f == NULL:
        Logger.debug('GL: glVertexAttrib4f is not available')
    if cgl.glVertexAttribPointer == NULL:
        Logger.debug('GL: glVertexAttribPointer is not available')
    if cgl.glViewport == NULL:
        Logger.debug('GL: glViewport is not available')



include '../../lib/sdl2.pxi'

import io
from kivy.logger import Logger
from libc.string cimport memset
from libc.stdlib cimport malloc

cdef int _is_init = 0

cdef struct SDL_RWops:
    long (* seek) (SDL_RWops * context, long offset,int whence)
    size_t(* read) ( SDL_RWops * context, void *ptr, size_t size, size_t maxnum)
    size_t(* write) (SDL_RWops * context, void *ptr,size_t size, size_t num)
    int (* close) (SDL_RWops * context)


cdef size_t rwops_bytesio_write(SDL_RWops *context, const void *ptr, size_t size, size_t num):
    cdef char *c_string = <char *>ptr
    byteio = <object>context.hidden.unknown.data1
    byteio.write(c_string[:size * num])
    return size * num


cdef int rwops_bytesio_close(SDL_RWops *context):
    byteio = <object>context.hidden.unknown.data1
    byteio.seek(0)


cdef SDL_RWops *rwops_bridge_to_bytesio(byteio):
    # works only for write.
    cdef SDL_RWops *rwops = SDL_AllocRW()
    rwops.hidden.unknown.data1 = <void *>byteio
    rwops.seek = NULL
    rwops.read = NULL
    rwops.write = &rwops_bytesio_write
    rwops.close =&rwops_bytesio_close
    return rwops


def init():
    global _is_init
    if _is_init:
        return

    cdef int ret
    for flags in (IMG_INIT_JPG, IMG_INIT_PNG, IMG_INIT_TIF, IMG_INIT_WEBP):
        ret = IMG_Init(flags)
        if ret & flags != flags:
            # FIXME replace flags by a good string
            Logger.error(
                'ImageSDL2: Failed to init required {} support'.format(flags))
            Logger.error('ImageSDL2: {}'.format(IMG_GetError()))

    _is_init = 1

def save(filename, w, h, pixelfmt, pixels, flipped, imagefmt, quality=90):
    cdef bytes c_filename = None
    cdef SDL_RWops *rwops
    if not isinstance(filename, io.BytesIO):
        c_filename = filename.encode('utf-8')
    cdef int pitch

    if pixelfmt == "rgb":
        pitch = w * 3
    elif pixelfmt == "rgba":
        pitch = w * 4
    else:
        raise Exception("IMG SDL2 supports only pixelfmt rgb and rgba")

    cdef int lng, top, bot
    cdef list rng
    if flipped:
        # if flipped upside down
        # switch bytes(array) to list
        pixels = list(pixels)
        lng = <int>len(pixels)
        rng = list(range(0, lng, pitch))

        while len(rng):
            try:
                top = rng.pop(0)
                bot = rng.pop()
            except IndexError:
                # odd height, single line remains, no swap
                break
            temp = pixels[top:top + pitch]
            pixels[top:top + pitch] = pixels[bot:bot + pitch]
            pixels[bot:bot + pitch] = temp

        # return to bytes(array) for sdl
        pixels = bytes(bytearray(pixels))

    cdef char *c_pixels = pixels
    cdef SDL_Surface *image = NULL

    if pixelfmt == "rgba":
        image = SDL_CreateRGBSurfaceFrom(
            c_pixels, w, h, 32, pitch,
            0x00000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)
    elif pixelfmt == "rgb":
        image = SDL_CreateRGBSurfaceFrom(
            c_pixels, w, h, 24, pitch,
            0x0000ff, 0x00ff00, 0xff0000, 0)

    if c_filename is not None:
        if imagefmt == "png":
            IMG_SavePNG(image, c_filename)
        elif imagefmt == "jpg":
            IMG_SaveJPG(image, c_filename, quality)
    else:
        rwops = rwops_bridge_to_bytesio(filename)
        if imagefmt == "png":
            IMG_SavePNG_RW(image, rwops, 1)
        elif imagefmt == "jpg":
            IMG_SaveJPG_RW(image, rwops, 1, quality)
        SDL_FreeRW(rwops)

    if image:
        SDL_FreeSurface(image)

# NOTE: This must be kept up to date with ImageData supported formats. If you
# add support for converting/uploading (for example) ARGB, you must ensure
# that it is returned unmodified below to avoid converting to RGB/RGBA.
#
# FIXME:
# - Some PNG48 images with binary transparency load incorrectly
# - (Some?) 8/16 bit grayscale PNGs load incorrectly (likely SDL_Image problem)
cdef load_from_surface(SDL_Surface *image):
    cdef SDL_Surface *image2 = NULL
    cdef SDL_Surface *fimage = NULL
    cdef Uint32 want_rgba = 0, want_bgra = 0, target_fmt = 0
    cdef int n = 0
    cdef bytes pixels

    if image == NULL:
        Logger.warn('ImageSDL2: load_from_surface() with NULL surface')
        return None

    # SDL 2.0.5 now has endian-agnostic 32-bit pixel formats like RGB24,
    # but we can't count on that yet (RGB24 is available since 2.0)
    if SDL_BYTEORDER == SDL_BIG_ENDIAN:
        want_rgba = SDL_PIXELFORMAT_RGBA8888
        want_bgra = SDL_PIXELFORMAT_BGRA8888
    else:
        want_rgba = SDL_PIXELFORMAT_ABGR8888
        want_bgra = SDL_PIXELFORMAT_ARGB8888

    # Output format (string) - supported by ImageData. If the surface is
    # already in a supported pixel format, no conversion is done.
    fmt = ''

    # 32-bit rgba and bgra can be used directly
    if image.format.format == want_rgba:
        fmt = 'rgba'
    elif image.format.format == want_bgra:
        fmt = 'bgra'

    # Alpha mask or colorkey must be converted to rgba
    elif image.format.Amask or SDL_GetColorKey(image, NULL) == 0:
        fmt = 'rgba'
        target_fmt = want_rgba

    # Palette with alpha is converted to rgba
    elif image.format.palette != NULL:
        for n in xrange(0, image.format.palette.ncolors):
            if image.format.palette.colors[n].a < 0xFF:
                fmt = 'rgba'
                target_fmt = want_rgba
                break

    # 24bpp RGB/BGR without colorkey can be used directly.
    elif image.format.format == SDL_PIXELFORMAT_RGB24:
        fmt = 'rgb'
    elif image.format.format == SDL_PIXELFORMAT_BGR24:
        fmt = 'bgr'

    # Everything else is converted to RGB
    if not fmt:
        fmt = 'rgb'
        target_fmt = SDL_PIXELFORMAT_RGB24

    # Convert if needed, and return a copy of the raw pixel data
    try:
        fimage = image
        if target_fmt != 0:
            with nogil:
                image2 = SDL_ConvertSurfaceFormat(image, target_fmt, 0)
            if image2 == NULL:
                Logger.warn('ImageSDL2: error converting {} to {}: {}'.format(
                        SDL_GetPixelFormatName(image.format.format),
                        SDL_GetPixelFormatName(target_fmt),
                        SDL_GetError()))
                return None
            else:
                fimage = image2

        pixels = (<char *>fimage.pixels)[:fimage.pitch * fimage.h]
        return (fimage.w, fimage.h, fmt, pixels, fimage.pitch)
    finally:
        if image2:
            SDL_FreeSurface(image2)


def load_from_filename(filename):
    cdef bytes c_filename = filename.encode('utf-8')
    cdef SDL_Surface *image = IMG_Load(c_filename)
    if image == NULL:
        return
    try:
        return load_from_surface(image)
    finally:
        if image:
            SDL_FreeSurface(image)

def load_from_memory(bytes data):
    cdef SDL_RWops *rw = NULL
    cdef SDL_Surface *image = NULL
    cdef char *c_data = data

    rw = SDL_RWFromMem(c_data, <int>len(data))
    if rw == NULL:
        return

    image = IMG_Load_RW(rw, 0)
    if image == NULL:
        return

    try:
        return load_from_surface(image)
    finally:
        if image:
            SDL_FreeSurface(image)
        if rw:
            SDL_FreeRW(rw)
